"""Fit by compressing chroma in LCh."""
from __future__ import annotations
from ..gamut import Fit, clip_channels
from ..cat import WHITES
from .. import util
import math
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color

WHITE = util.xy_to_xyz(WHITES['2deg']['D65'])
BLACK = [0, 0, 0]


class LChChroma(Fit):
    """
    LCh chroma gamut mapping class.

    Adjust chroma (using binary search).
    This helps preserve the other attributes of the color.
    Compress chroma until we are are right at the JND edge while still out of gamut.
    Raise the lower chroma bound while we are in gamut or outside of gamut but still under the JND.
    Lower the upper chroma bound anytime we are out of gamut and above the JND.
    Too far under the JND we'll reduce chroma too aggressively.

    This is the same as the CSS algorithm as described here: https://www.w3.org/TR/css-color-4/#binsearch.
    There are some small adjustments to handle HDR colors as the CSS algorithm assumes SDR color spaces.
    Additionally, this uses LCh instead of OkLCh, but we also offer a derived version that uses OkLCh.
    """

    NAME = "lch-chroma"

    EPSILON = 0.1
    LIMIT = 2.0
    DE = "2000"
    DE_OPTIONS = {}  # type: dict[str, Any]
    SPACE = "lch-d65"
    MIN_LIGHTNESS = 0
    MAX_LIGHTNESS = 100
    MIN_CONVERGENCE = 0.0001

    def fit(self, color: Color, **kwargs: Any) -> None:
        """Gamut mapping via CIELCh chroma."""

        space = color.space()
        mapcolor = color.convert(self.SPACE, norm=False)
        l, c = mapcolor._space.indexes()[:2]  # type: ignore[attr-defined]
        lightness = mapcolor[l]
        sdr = color._space.DYNAMIC_RANGE == 'sdr'

        # Return white or black if lightness is out of dynamic range for lightness.
        # Extreme light case only applies to SDR, but dark case applies to all ranges.
        if sdr and (lightness >= self.MAX_LIGHTNESS or math.isclose(lightness, self.MAX_LIGHTNESS, abs_tol=1e-6)):
            clip_channels(color.update('xyz-d65', WHITE, mapcolor[-1]))
            return
        elif lightness <= self.MIN_LIGHTNESS:
            clip_channels(color.update('xyz-d65', BLACK, mapcolor[-1]))
            return

        # Set initial chroma boundaries
        low = 0.0
        high = mapcolor[c]
        clip_channels(color._hotswap(mapcolor.convert(space, norm=False)))

        # Adjust chroma if we are not under the JND yet.
        if mapcolor.delta_e(color, method=self.DE, **self.DE_OPTIONS) >= self.LIMIT:
            # Perform "in gamut" checks until we know our lower bound is no longer in gamut.
            lower_in_gamut = True

            # If high and low get too close to converging,
            # we need to quit in order to prevent infinite looping.
            while (high - low) > self.MIN_CONVERGENCE:
                mapcolor[c] = (high + low) * 0.5

                # Avoid doing expensive delta E checks if in gamut
                if lower_in_gamut and mapcolor.in_gamut(space, tolerance=0):
                    low = mapcolor[c]
                else:
                    clip_channels(color._hotswap(mapcolor.convert(space, norm=False)))
                    de = mapcolor.delta_e(color, method=self.DE, **self.DE_OPTIONS)
                    if de < self.LIMIT:
                        # Kick out as soon as we are close enough to the JND.
                        # Too far below and we may reduce chroma too aggressively.
                        if (self.LIMIT - de) < self.EPSILON:
                            break

                        # Our lower bound is now out of gamut, so all future searches are
                        # guaranteed to be out of gamut. Now we just want to focus on tuning
                        # chroma to get as close to the JND as possible.
                        if lower_in_gamut:
                            lower_in_gamut = False
                        low = mapcolor[c]
                    else:
                        # We are still outside the gamut and outside the JND
                        high = mapcolor[c]
        color.normalize()



"""Continuous interpolation."""
from __future__ import annotations
from .. import algebra as alg
from ..interpolate import Interpolator, Interpolate
from ..types import Vector
from typing import Callable, Mapping, Sequence, Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class InterpolatorContinuous(Interpolator):
    """Interpolate with continuous piecewise."""

    def handle_undefined(self) -> None:
        """
        Handle null values.

        Resolve any undefined alpha values and apply premultiplication if necessary.

        Additionally, any undefined value have a new control point generated via
        linear interpolation. This is the only approach to provide a non-bias, non-breaking
        way to handle things like achromatic hues in a cylindrical space. It also balances
        non cylindrical values. Since the B-spline needs a a continual path and since we
        have a sliding window that takes into account 4 points at a time, we must consider
        a more broad context than what is done in piecewise linear.
        """

        coords = self.coordinates

        # Process each set of coordinates
        alpha = len(coords[0]) - 1
        for i in range(len(coords[0])):
            backfill = None
            last = None

            # Process a specific channel for all coordinates sets
            for x in range(1, len(coords)):
                c1, c2 = coords[x - 1:x + 1]
                a, b = c1[i], c2[i]
                a_nan, b_nan = alg.is_nan(a), alg.is_nan(b)

                # Two good values, store the last good value and continue
                if not a_nan and not b_nan:
                    if self.premultiplied and i == alpha:
                        self.premultiply(c1)
                        self.premultiply(c2)
                    last = b
                    continue

                # Found a gap
                if a_nan:
                    # First color starts an undefined gap
                    if backfill is None:
                        backfill = x - 1

                    # Gap continues
                    if b_nan:
                        continue

                    if self.premultiplied and i == alpha:
                        self.premultiply(c2)

                    # Generate new control points for the undefined value. Use linear
                    # interpolation if two known values bookend the undefined gap,
                    # else just backfill the current known value.
                    point = 1 / (x - backfill + 1)
                    for e, c in enumerate(coords[backfill:x], 1):
                        p = alg.lerp(last, b, point * e) if last is not None else b
                        c[i] = p

                        # We just filled an alpha hole, premultiply the coordinates
                        if self.premultiplied and i == alpha:
                            self.premultiply(c)

                    backfill = None
                    last = b
                else:
                    # Started a new gap after a good value
                    # This always starts a new gap and never finishes one
                    if backfill is None:
                        backfill = x

                    if self.premultiplied and i == alpha:
                        self.premultiply(c1)
                    last = a

            # Replace all undefined values that occurred prior to
            # finding the current defined value that have not been backfilled
            if backfill is not None and last is not None:
                for c in coords[backfill:]:
                    c[i] = last

                    # We just filled an alpha hole, premultiply the coordinates
                    if self.premultiplied and i == alpha:
                        self.premultiply(c)

    def setup(self) -> None:
        """Optional setup."""

        # Process undefined values
        self.handle_undefined()

    def interpolate(
        self,
        point: float,
        index: int
    ) -> Vector:
        """Interpolate."""

        # Interpolate between the values of the two colors for each channel.
        channels = []
        idx = index - 2 if index == self.length else index - 1

        for i, values in enumerate(zip(*self.coordinates[idx:idx + 2])):
            a, b = values

            # Interpolate
            value = alg.lerp(a, b, self.ease(point, i))
            channels.append(value)

        return channels


class Continuous(Interpolate):
    """Continuous interpolation plugin."""

    NAME = "continuous"

    def interpolator(
        self,
        coordinates: list[Vector],
        channel_names: Sequence[str],
        create: type[Color],
        easings: list[Callable[..., float] | None],
        stops: dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: list[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ) -> Interpolator:
        """Return the B-spline interpolator."""

        return InterpolatorContinuous(
            coordinates,
            channel_names,
            create,
            easings,
            stops,
            space,
            out_space,
            progress,
            premultiplied,
            extrapolate,
            domain,
            padding,
            **kwargs
        )



"""Test CMY."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestCMY(util.ColorAssertsPyTest):
    """Test CMY."""

    COLORS = [
        ('red', 'color(--cmy 0 1 1)'),
        ('orange', 'color(--cmy 0 0.35294 1)'),
        ('yellow', 'color(--cmy 0 0 1)'),
        ('green', 'color(--cmy 1 0.49804 1)'),
        ('blue', 'color(--cmy 1 1 0)'),
        ('indigo', 'color(--cmy 0.70588 1 0.4902)'),
        ('violet', 'color(--cmy 0.06667 0.4902 0.06667)'),
        ('white', 'color(--cmy 0 0 0)'),
        ('gray', 'color(--cmy 0.49804 0.49804 0.49804)'),
        ('black', 'color(--cmy 1 1 1)'),
        # Test CSS color
        ('color(--cmy 0 0.50196 0)', 'color(--cmy 0 0.50196 0)'),
        ('color(--cmy 0 0.50196 0 / 0.5)', 'color(--cmy 0 0.50196 0 / 0.5)'),
        ('color(--cmy 50% 50% 50% / 50%)', 'color(--cmy 0.5 0.5 0.5 / 0.5)'),
        ('color(--cmy none none none / none)', 'color(--cmy none none none / none)'),
        # Test range
        ('color(--cmy 0% 0% 0%)', 'color(--cmy 0 0 0)'),
        ('color(--cmy 100% 100% 100%)', 'color(--cmy 1 1 1)'),
        ('color(--cmy -100% -100% -100%)', 'color(--cmy -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_cmy_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('cmy'), Color(color2))


class TestCMYSerialize(util.ColorAssertsPyTest):
    """Test CMY serialization."""

    COLORS = [
        # Test color
        ('color(--cmy 0 0.3 0.75 / 0.5)', {}, 'color(--cmy 0 0.3 0.75 / 0.5)'),
        # Test alpha
        ('color(--cmy 0 0.3 0.75)', {'alpha': True}, 'color(--cmy 0 0.3 0.75 / 1)'),
        ('color(--cmy 0 0.3 0.75 / 0.5)', {'alpha': False}, 'color(--cmy 0 0.3 0.75)'),
        # Test None
        ('color(--cmy none 0.3 0.75)', {}, 'color(--cmy 0 0.3 0.75)'),
        ('color(--cmy none 0.3 0.75)', {'none': True}, 'color(--cmy none 0.3 0.75)'),
        # Test Fit
        ('color(--cmy 1.2 0.2 0)', {}, 'color(--cmy 1 0.2 0)'),
        ('color(--cmy 1.2 0.2 0)', {'fit': False}, 'color(--cmy 1.2 0.2 0)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestCMYPoperties(util.ColorAsserts, unittest.TestCase):
    """Test CMY."""

    def test_c(self):
        """Test `c`."""

        c = Color('color(--cmy 0 0 1)')
        self.assertEqual(c['c'], 0)
        c['c'] = 0.2
        self.assertEqual(c['c'], 0.2)

    def test_m(self):
        """Test `m`."""

        c = Color('color(--cmy 0 0 1)')
        self.assertEqual(c['m'], 0)
        c['m'] = 0.1
        self.assertEqual(c['m'], 0.1)

    def test_y(self):
        """Test `y`."""

        c = Color('color(--cmy 0 0 1)')
        self.assertEqual(c['y'], 1)
        c['y'] = 0.1
        self.assertEqual(c['y'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--cmy 0 0 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('cmy', [0.3, 0.3, 0.3]).is_achromatic(), True)
        self.assertEqual(Color('cmy', [0.3, 0.3, 0.30000001]).is_achromatic(), True)
        self.assertEqual(Color('cmy', [0.3, 0.4, 0.3]).is_achromatic(), False)
        self.assertEqual(Color('cmy', [0.3, 0.3, NaN]).is_achromatic(), False)



"""Test chromatic adaptation."""
from coloraide.everything import ColorAll as Color
from coloraide import cat
from coloraide import util as cutil
from . import util
import pytest


class TestCAT(util.ColorAssertsPyTest):
    """Test CAT."""

    @pytest.mark.parametrize(
        'method,cat1,cat2',
        [
            ('bradford', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('von-kries', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('xyz-scaling', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('cat02', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('cmccat97', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('sharp', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('cmccat2000', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50']),
            ('cat16', cat.WHITES['2deg']['D65'], cat.WHITES['2deg']['D50'])

        ]
    )
    def test_cat(self, method, cat1, cat2):
        """Test CAT methods."""

        c1 = Color('white').convert('xyz-d65')
        c2 = Color('xyz-d50', c1.chromatic_adaptation(cat1, cat2, c1[:-1], method=method))
        c3 = Color('xyz-d65', c2.chromatic_adaptation(cat2, cat1, c2[:-1], method=method))
        self.assertColorEqual(c1, c3)
        self.assertColorEqual(c2, Color('xyz-d50', cutil.xy_to_xyz(cat2)))



"""Test UCS 1960."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
import pytest


class TestUCS(util.ColorAssertsPyTest):
    """Test UCS 1960."""

    COLORS = [
        ('red', 'color(--ucs 0.27493 0.21264 0.12243)'),
        ('orange', 'color(--ucs 0.36462 0.48173 0.48122)'),
        ('yellow', 'color(--ucs 0.51332 0.92781 1.076)'),
        ('green', 'color(--ucs 0.05146 0.15438 0.20584)'),
        ('blue', 'color(--ucs 0.12032 0.07219 0.49331)'),
        ('indigo', 'color(--ucs 0.0462 0.03108 0.11874)'),
        ('violet', 'color(--ucs 0.39115 0.40317 0.73932)'),
        ('white', 'color(--ucs 0.63364 1 1.5693)'),
        ('gray', 'color(--ucs 0.13678 0.21586 0.33875)'),
        ('black', 'color(--ucs 0 0 0)'),
        # Test color
        ('color(--ucs 0.5 0.1 0.1)', 'color(--ucs 0.5 0.1 0.1)'),
        ('color(--ucs 0.5 0.1 0.1 / 0.5)', 'color(--ucs 0.5 0.1 0.1 / 0.5)'),
        ('color(--ucs 50% 50% 50% / 50%)', 'color(--ucs 0.5 0.5 0.5 / 0.5)'),
        ('color(--ucs none none none / none)', 'color(--ucs none none none / none)'),
        # Test percent ranges
        ('color(--ucs 0% 0% 0%)', 'color(--ucs 0 0 0)'),
        ('color(--ucs 100% 100% 100%)', 'color(--ucs 1 1 1)'),
        ('color(--ucs -100% -100% -100%)', 'color(--ucs -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_ucs_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('ucs'), Color(color2))


class TestUCSSerialize(util.ColorAssertsPyTest):
    """Test UCS serialization."""

    COLORS = [
        # Test color
        ('color(--ucs 0.75 0.1 0.1 / 0.5)', {}, 'color(--ucs 0.75 0.1 0.1 / 0.5)'),
        # Test alpha
        ('color(--ucs 0.75 0.1 0.1)', {'alpha': True}, 'color(--ucs 0.75 0.1 0.1 / 1)'),
        ('color(--ucs 0.75 0.1 0.1 / 0.5)', {'alpha': False}, 'color(--ucs 0.75 0.1 0.1)'),
        # Test None
        ('color(--ucs none 0.1 0.1)', {}, 'color(--ucs 0 0.1 0.1)'),
        ('color(--ucs none 0.1 0.1)', {'none': True}, 'color(--ucs none 0.1 0.1)'),
        # Test Fit
        ('color(--ucs 0.75 1.2 0.1)', {}, 'color(--ucs 0.75 1.2 0.1)'),
        ('color(--ucs 0.75 1.2 0.1)', {'fit': False}, 'color(--ucs 0.75 1.2 0.1)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestUCSPoperties(util.ColorAsserts, unittest.TestCase):
    """Test UCS."""

    def test_u(self):
        """Test `u`."""

        c = Color('color(--ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['u'], 0.51332)
        c['u'] = 0.2
        self.assertEqual(c['u'], 0.2)

    def test_v(self):
        """Test `v`."""

        c = Color('color(--ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['v'], 0.92781)
        c['v'] = 0.1
        self.assertEqual(c['v'], 0.1)

    def test_w(self):
        """Test `w`."""

        c = Color('color(--ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['w'], 1.076)
        c['w'] = 0.1
        self.assertEqual(c['w'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Chromatic adaptation transforms."""
from __future__ import annotations
from . import util
from abc import ABCMeta, abstractmethod
from . import algebra as alg
import functools
from .types import Matrix, VectorLike, Vector, Plugin
from typing import Any  # noqa: F401

# From CIE 2004 Colorimetry T.3 and T.8
# B from https://en.wikipedia.org/wiki/Standard_illuminant#White_point
WHITES = {
    "2deg": {
        "A": (0.44758, 0.40745),
        "B": (0.34842, 0.35161),
        "C": (0.31006, 0.31616),
        "D50": (0.34570, 0.35850),  # Use 4 digits like everyone
        "D55": (0.33243, 0.34744),
        "D65": (0.31270, 0.32900),  # Use 4 digits like everyone
        "D75": (0.29903, 0.31488),
        "E": (1 / 3, 1 / 3),
        "F2": (0.37210, 0.37510),
        "F7": (0.31290, 0.32920),
        "F11": (0.38050, 0.37690)
    },

    "10deg": {
        "A": (0.45117, 0.40594),
        "B": (0.34980, 0.35270),
        "C": (0.31039, 0.31905),
        "D50": (0.34773, 0.35952),
        "D55": (0.33412, 0.34877),
        "D65": (0.31382, 0.33100),
        "D75": (0.29968, 0.31740),
        "E": (1 / 3, 1 / 3),
        "F2": (0.37925, 0.36733),
        "F3": (0.41761, 0.38324),
        "F11": (0.38541, 0.37123)
    }
}


def calc_adaptation_matrices(
    w1: tuple[float, float],
    w2: tuple[float, float],
    m: Matrix,
) -> tuple[Matrix, Matrix]:
    """
    Get the von Kries based adaptation matrix based on the method and illuminants.

    Since these calculated matrices are cached, this greatly reduces
    performance hit as the initial matrices only have to be calculated
    once for a given pair of white points and CAT.

    Granted, we are currently, capped at 20 in the cache, but the average user
    isn't going to be swapping between over 20 methods and white points in a
    short period of time. We could always increase the cache if necessary.
    """

    first = alg.dot(m, util.xy_to_xyz(w1), dims=alg.D2_D1)
    second = alg.dot(m, util.xy_to_xyz(w2), dims=alg.D2_D1)
    m2 = alg.diag(alg.divide(first, second, dims=alg.D1))
    adapt = alg.multi_dot([alg.inv(m), m2, m])  # type: Any

    return adapt, alg.inv(adapt)


class CAT(Plugin, metaclass=ABCMeta):
    """Chromatic adaptation."""

    NAME = ''

    @abstractmethod
    def adapt(self, w1: tuple[float, float], w2: tuple[float, float], xyz: VectorLike) -> Vector:
        """Adapt a given XYZ color using the provided white points."""


class VonKries(CAT):
    """
    Von Kries CAT.

    http://brucelindbloom.com/Eqn_ChromAdapt.html
    https://hrcak.srce.hr/file/95370
    """

    NAME = 'von-kries'

    MATRIX = [
        [0.4002400, 0.7076000, -0.0808100],
        [-0.2263000, 1.1653200, 0.0457000],
        [0.0000000, 0.0000000, 0.9182200]
    ]  # type: Matrix

    @classmethod
    @functools.lru_cache(maxsize=20)
    def get_adaptation_matrices(
        cls: type[VonKries],
        w1: tuple[float, float],
        w2: tuple[float, float]
    ) -> tuple[Matrix, Matrix]:
        """Get the adaptation matrices."""

        return calc_adaptation_matrices(w1, w2, cls.MATRIX)

    def adapt(self, w1: tuple[float, float], w2: tuple[float, float], xyz: VectorLike) -> Vector:
        """Adapt a given XYZ color using the provided white points."""

        # We are already using the correct white point
        if w1 == w2:
            return list(xyz)

        a, b = sorted([w1, w2])
        m, mi = self.get_adaptation_matrices(a, b)
        return alg.dot(mi if a != w2 else m, xyz, dims=alg.D2_D1)


class Bradford(VonKries):
    """
    Bradford CAT.

    http://brucelindbloom.com/Eqn_ChromAdapt.html
    https://hrcak.srce.hr/file/95370
    """

    NAME = "bradford"

    MATRIX = [
        [0.8951000, 0.2664000, -0.1614000],
        [-0.7502000, 1.7135000, 0.0367000],
        [0.0389000, -0.0685000, 1.0296000]
    ]


class XYZScaling(VonKries):
    """
    XYZ Scaling CAT.

    http://brucelindbloom.com/Eqn_ChromAdapt.html
    https://hrcak.srce.hr/file/95370
    """

    NAME = "xyz-scaling"

    MATRIX = [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ]


class CAT02(VonKries):
    """
    CAT02 CAT.

    https://en.wikipedia.org/wiki/CIECAM02#CAT02
    """

    NAME = "cat02"

    MATRIX = [
        [0.7328000, 0.4296000, -0.1624000],
        [-0.7036000, 1.6975000, 0.0061000],
        [0.0030000, 0.0136000, 0.9834000]
    ]


class CMCCAT97(VonKries):
    """
    CMCCAT97 CAT.

    https://hrcak.srce.hr/file/95370
    """

    NAME = "cmccat97"

    MATRIX = [
        [0.8951000, -0.7502000, 0.0389000],
        [0.2664000, 1.7135000, 0.0685000],
        [-0.1614000, 0.0367000, 1.0296000],
    ]


class Sharp(VonKries):
    """
    Sharp CAT.

    https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.14.918&rep=rep1&type=pdf
    """

    NAME = "sharp"

    MATRIX = [
        [1.2694000, -0.0988000, -0.1706000],
        [-0.8364000, 1.8006000, 0.0357000],
        [0.0297000, -0.0315000, 1.0018000]
    ]


class CMCCAT2000(VonKries):
    """
    CMCCAT2000 CAT.

    https://hrcak.srce.hr/file/95370
    https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.14.918&rep=rep1&type=pdf
    """

    NAME = 'cmccat2000'

    MATRIX = [
        [0.7982000, 0.3389000, -0.1371000],
        [-0.5918000, 1.5512000, 0.0406000],
        [0.0008000, 0.0239000, 0.9753000]
    ]


class CAT16(VonKries):
    """
    CAT16 CAT.

    https://arxiv.org/pdf/1802.06067.pdf
    """

    NAME = "cat16"

    MATRIX = [
        [0.401288, 0.650173, -0.051461],
        [-0.250268, 1.204414, 0.045854],
        [-0.002079, 0.048952, 0.953127]
    ]



"""LCh class."""
from __future__ import annotations
from ...spaces import Space, LChish
from ...cat import WHITES
from ...channels import Channel, FLG_ANGLE, FLG_OPT_PERCENT
from ... import util
import math
from ...types import Vector

ACHROMATIC_THRESHOLD = 1e-4


def lab_to_lch(lab: Vector) -> Vector:
    """Lab to LCh."""

    l, a, b = lab

    c = math.sqrt(a ** 2 + b ** 2)
    h = math.degrees(math.atan2(b, a))

    return [l, c, util.constrain_hue(h)]


def lch_to_lab(lch: Vector) -> Vector:
    """LCh to Lab."""

    l, c, h = lch

    return [
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    ]


class LCh(LChish, Space):
    """LCh class."""

    BASE = "lab"
    NAME = "lch"
    SERIALIZE = ("--lch",)
    CHANNELS = (
        Channel("l", 0.0, 100.0, flags=FLG_OPT_PERCENT),
        Channel("c", 0.0, 150.0, limit=(0.0, None), flags=FLG_OPT_PERCENT),
        Channel("h", 0.0, 360.0, flags=FLG_ANGLE)
    )
    CHANNEL_ALIASES = {
        "lightness": "l",
        "chroma": "c",
        "hue": "h"
    }
    WHITE = WHITES['2deg']['D50']

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return coords[1] < ACHROMATIC_THRESHOLD

    def to_base(self, coords: Vector) -> Vector:
        """To Lab from LCh."""

        return lch_to_lab(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From Lab to LCh."""

        return lab_to_lch(coords)



"""Display-p3 color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB, lin_srgb, gam_srgb
from ..types import Vector


class DisplayP3(sRGB):
    """Display-p3 class."""

    BASE = "display-p3-linear"
    NAME = "display-p3"
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Display P3."""

        return lin_srgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Display P3."""

        return gam_srgb(coords)



"""Tools for dynamic achromatic response."""
from __future__ import annotations
from .. import algebra as alg
import bisect
from typing import Any
from ..types import Vector, VectorLike
from abc import ABCMeta, abstractmethod
import math


class Achromatic(metaclass=ABCMeta):
    """Calculate a spline that follows a color's achromatic response."""

    L_IDX = 0
    C_IDX = 1
    H_IDX = 2

    def __init__(
        self,
        data: list[Vector] | None = None,
        threshold_upper: float = float('inf'),
        threshold_lower: float = float('inf'),
        threshold_cutoff: float = float('inf'),
        spline: str = 'linear',
        mirror: bool = False,
        **kwargs: Any
    ) -> None:
        """
        Initialize.

        `tuning`: Either a dictionary of `low`, `mid`, and `high`, each specifying a start, end, step, and scale
            used to build a portion of the spline representing the achromatic response or a list containing the
            the pre-calculated spline points.
        `threshold_upper`: threshold above achromatic curve.
        `threshold_lower`: threshold below achromatic curve.
        `threshold_cutoff`: threshold of chroma above which we will assume colors are not achromatic.
        `spline`: spline type.
        `mirror`: Mirror response across lightness axis at zero.
        """

        self.mirror = mirror
        self.threshold_upper = threshold_upper
        self.threshold_lower = threshold_lower
        self.threshold_cutoff = threshold_cutoff

        self.domain = []  # type: list[float]
        self.min_colorfulness = 1e10
        self.min_lightness = 1e10
        self.spline_type = spline

        # Create a spline that maps the achromatic range for the SDR range
        if data is not None:
            self.setup_achromatic_response(data, **kwargs)

    def dump(self) -> list[VectorLike] | None:  # pragma: no cover
        """Dump data points."""

        if self.spline_type == 'linear':
            return list(zip(*self.spline.points))
        else:
            # Strip off the data points used to coerce the spline through the end.
            return list(zip(*self.spline.points))[1:-1]

    @abstractmethod
    def convert(self, coords: Vector, **kwargs: Any) -> Vector:
        """Convert an sRGB color to the desired space."""

    def calc_achromatic_response(
        self,
        parameters: list[tuple[int, int, int, float]],
        **kwargs: Any
    ) -> None:  # pragma: no cover
        """
        Calculate the achromatic response.

        Used to precalculate the best response.
        """

        points = []  # type: list[list[float]]
        for segment in parameters:
            start, end, step, scale = segment
            for p in range(start, end, step):
                color = self.convert([p / scale] * 3, **kwargs)
                l, c, h = color[self.L_IDX], color[self.C_IDX], color[self.H_IDX]
                if l < self.min_lightness:
                    self.min_lightness = l
                if c < self.min_colorfulness:
                    self.min_colorfulness = c
                self.domain.append(l)
                points.append([l, c, h % 360])
        self.spline = alg.interpolate(points, method=self.spline_type)
        self.hue = self.convert([1] * 3, **kwargs)[self.H_IDX] % 360
        self.ihue = (self.hue - 180) % 360

    def setup_achromatic_response(
        self,
        tuning: list[Vector],
        **kwargs: Any
    ) -> None:
        """Setup the achromatic response."""

        points = []  # type: list[list[float]]
        for entry in tuning:
            l, c, h = entry
            if l < self.min_lightness:
                self.min_lightness = l
            if c < self.min_colorfulness:
                self.min_colorfulness = c
            points.append([l, c, h])
            self.domain.append(l)
        self.spline = alg.interpolate(points, method=self.spline_type)
        self.hue = self.convert([1] * 3, **kwargs)[self.H_IDX] % 360
        self.ihue = (self.hue - 180) % 360

    def scale(self, point: float) -> float:
        """Scale the lightness to match the range."""

        if point <= self.domain[0]:
            point = (point - self.domain[0]) / (self.domain[-1] - self.domain[0])
        elif point >= self.domain[-1]:
            point = 1.0 + (point - self.domain[-1]) / (self.domain[-1] - self.domain[0])
        else:
            regions = len(self.domain) - 1
            size = (1 / regions)
            index = 0
            adjusted = 0.0
            index = bisect.bisect(self.domain, point) - 1
            a, b = self.domain[index:index + 2]
            l = b - a
            adjusted = ((point - a) / l) if l else 0.0
            point = size * index + (adjusted * size)
        return point

    def get_ideal_chroma(self, l: float) -> float:
        """Get the ideal chroma."""

        if math.isnan(l):
            return 0.0

        elif self.mirror and l < 0.0:
            return self.spline(self.scale(abs(l)))[1]

        return self.spline(self.scale(l))[1]

    def get_ideal_hue(self, l: float) -> float:
        """Get the ideal chroma."""

        if math.isnan(l):
            return 0.0

        elif self.mirror and l < 0.0:
            return (self.spline(self.scale(abs(l)))[2] - 180) % 360

        return self.spline(self.scale(l))[2]

    def get_ideal_ab(self, l: float) -> tuple[float, float]:
        """Get the ideal rectangular form of chroma and hue, the components a and b."""

        if math.isnan(l):
            return 0.0, 0.0

        return alg.polar_to_rect(self.get_ideal_chroma(l), self.get_ideal_hue(l))

    def test(self, l: float, c: float, h: float) -> bool:
        """Test if the current color is achromatic."""

        # If colorfulness is past this limit, we'd have to have a lightness
        # so high, that our test has already broken down.
        if c > self.threshold_cutoff or (not self.mirror and l < 0.0):
            return False

        # If we are higher than 1, we are extrapolating;
        # otherwise, use the spline.
        flip = self.mirror and l < 0.0
        la = abs(l)
        point = self.scale(la if flip else l)
        if la < self.min_lightness and c < self.min_colorfulness:  # pragma: no cover
            return True
        else:
            c2, h2 = self.spline(point)[1:]
            if flip:
                h2 = (h2 - 180) % 360
        diff = c2 - c
        hdiff = abs(h % 360 - h2)
        if hdiff > 180:  # pragma: no cover
            hdiff = 360 - hdiff
        return (
            ((diff >= 0 and diff < self.threshold_upper) or (diff < 0 and abs(diff) < self.threshold_lower)) and
            (c2 < 1e-5 or hdiff < 0.01)
        )



"""Test Linear Rec. 2020."""
import unittest
from . import util
from coloraide import Color
import pytest


class TestLinearRec2020(util.ColorAssertsPyTest):
    """Test Linear Rec. 2020."""

    COLORS = [
        ('red', 'color(--rec2020-linear 0.6274 0.0691 0.01639)'),
        ('orange', 'color(--rec2020-linear 0.7513 0.41509 0.04951)'),
        ('yellow', 'color(--rec2020-linear 0.95669 0.98864 0.1044)'),
        ('green', 'color(--rec2020-linear 0.07108 0.19849 0.019)'),
        ('blue', 'color(--rec2020-linear 0.04331 0.01136 0.8956)'),
        ('indigo', 'color(--rec2020-linear 0.05381 0.0074 0.20108)'),
        ('violet', 'color(--rec2020-linear 0.64696 0.27406 0.79939)'),
        ('white', 'color(--rec2020-linear 1 1 1)'),
        ('gray', 'color(--rec2020-linear 0.21586 0.21586 0.21586)'),
        ('black', 'color(--rec2020-linear 0 0 0)'),
        # Test CSS color
        ('color(--rec2020-linear 0 0.50196 0)', 'color(--rec2020-linear 0 0.50196 0)'),
        ('color(--rec2020-linear 0 0.50196 0 / 0.5)', 'color(--rec2020-linear 0 0.50196 0 / 0.5)'),
        ('color(--rec2020-linear 50% 50% 50% / 50%)', 'color(--rec2020-linear 0.5 0.5 0.5 / 0.5)'),
        ('color(--rec2020-linear none none none / none)', 'color(--rec2020-linear none none none / none)'),
        # Test range
        ('color(--rec2020-linear 0% 0% 0%)', 'color(--rec2020-linear 0 0 0)'),
        ('color(--rec2020-linear 100% 100% 100%)', 'color(--rec2020-linear 1 1 1)'),
        ('color(--rec2020-linear -100% -100% -100%)', 'color(--rec2020-linear -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('rec2020-linear'), Color(color2))


class TestLinearRec2020Serialize(util.ColorAssertsPyTest):
    """Test Linear Rec. 2020 serialization."""

    COLORS = [
        # Test color
        ('color(--rec2020-linear 0 0.3 0.75 / 0.5)', {}, 'color(--rec2020-linear 0 0.3 0.75 / 0.5)'),
        # Test alpha
        ('color(--rec2020-linear 0 0.3 0.75)', {'alpha': True}, 'color(--rec2020-linear 0 0.3 0.75 / 1)'),
        ('color(--rec2020-linear 0 0.3 0.75 / 0.5)', {'alpha': False}, 'color(--rec2020-linear 0 0.3 0.75)'),
        # Test None
        ('color(--rec2020-linear none 0.3 0.75)', {}, 'color(--rec2020-linear 0 0.3 0.75)'),
        ('color(--rec2020-linear none 0.3 0.75)', {'none': True}, 'color(--rec2020-linear none 0.3 0.75)'),
        # Test Fit
        ('color(--rec2020-linear 1.2 0.2 0)', {}, 'color(--rec2020-linear 1 0.24165 0.01231)'),
        ('color(--rec2020-linear 1.2 0.2 0)', {'fit': False}, 'color(--rec2020-linear 1.2 0.2 0)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestLinearRec2020Poperties(util.ColorAsserts, unittest.TestCase):
    """Test Linear Rec. 2020 properties."""

    def test_r(self):
        """Test `r`."""

        c = Color('color(--rec2020-linear 0.1 0.2 0.3)')
        self.assertEqual(c['r'], 0.1)
        c['r'] = 0.2
        self.assertEqual(c['r'], 0.2)

    def test_g(self):
        """Test `g`."""

        c = Color('color(--rec2020-linear 0.1 0.2 0.3)')
        self.assertEqual(c['g'], 0.2)
        c['g'] = 0.1
        self.assertEqual(c['g'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--rec2020-linear 0.1 0.2 0.3)')
        self.assertEqual(c['b'], 0.3)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--rec2020-linear 0.1 0.2 0.3)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""
CAM16 class.

https://www.imaging.org/site/PDFS/Papers/2000/PICS-0-81/1611.pdf
https://observablehq.com/@jrus/cam16
https://arxiv.org/abs/1802.06067
https://doi.org/10.1002/col.22131
"""
from __future__ import annotations
import math
from .cam16_jmh import CAM16JMh
from ..spaces import Space, Labish
from ..cat import WHITES
from ..channels import Channel, FLG_MIRROR_PERCENT
from .. import util
from ..types import Vector
from .. import algebra as alg


def cam16_jmh_to_cam16_jab(jmh: Vector) -> Vector:
    """Translate a CAM16 JMh to Jab of the same viewing conditions."""

    J, M, h = jmh
    return [
        J,
        M * math.cos(math.radians(h)),
        M * math.sin(math.radians(h))
    ]


def cam16_jab_to_cam16_jmh(jab: Vector) -> Vector:
    """Translate a CAM16 Jab to JMh of the same viewing conditions."""

    J, a, b = jab
    M = math.sqrt(a ** 2 + b ** 2)
    h = math.degrees(math.atan2(b, a))

    return [J, M, util.constrain_hue(h)]


class CAM16(Labish, Space):
    """CAM16 class (Jab)."""

    BASE = "cam16-jmh"
    NAME = "cam16"
    SERIALIZE = ("--cam16",)
    CHANNELS = (
        Channel("j", 0.0, 100.0, limit=(0.0, None)),
        Channel("a", -90.0, 90.0, flags=FLG_MIRROR_PERCENT),
        Channel("b", -90.0, 90.0, flags=FLG_MIRROR_PERCENT)
    )
    CHANNEL_ALIASES = {
        "lightness": "j"
    }
    WHITE = WHITES['2deg']['D65']
    # Use the same environment as CAM16JMh
    ENV = CAM16JMh.ENV
    ACHROMATIC = CAM16JMh.ACHROMATIC

    def resolve_channel(self, index: int, coords: Vector) -> float:
        """Resolve channels."""

        if index in (1, 2):
            if not math.isnan(coords[index]):
                return coords[index]

            return self.ACHROMATIC.get_ideal_ab(coords[0])[index - 1]

        value = coords[index]
        return self.channels[index].nans if math.isnan(value) else value

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        m, h = alg.rect_to_polar(coords[1], coords[2])
        return coords[0] == 0.0 or self.ACHROMATIC.test(coords[0], m, h)

    def to_base(self, coords: Vector) -> Vector:
        """To CAM16 JMh from CAM16."""

        return cam16_jab_to_cam16_jmh(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From CAM16 JMh to CAM16."""

        return cam16_jmh_to_cam16_jab(coords)



"""
Calculate CCT with Robertson 1968 method.

Uses Robertson 1968 method.

- https://en.wikipedia.org/wiki/Correlated_color_temperature#Robertson's_method
- http://www.brucelindbloom.com/index.html?Math.html
"""
from __future__ import annotations
import math
from . import planck
from .. import algebra as alg
from .. import util
from .. import cat
from .. import cmfs
from ..temperature import CCT
from ..types import Vector, VectorLike
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color

# Original 31 mired points 0 - 600
MIRED_ORIGINAL = tuple(range(0, 100, 10)) + tuple(range(100, 601, 25))
# Extended 16 mired points 625 - 1000
MIRED_EXTENDED = MIRED_ORIGINAL + tuple(range(625, 1001, 25))


class Robertson1968(CCT):
    """Delta E plugin class."""

    NAME = 'robertson-1968'
    CHROMATICITY = 'uv-1960'

    def __init__(
        self,
        cmfs: dict[int, tuple[float, float, float]] = cmfs.CIE_1931_2DEG,
        white: VectorLike = cat.WHITES['2deg']['D65'],
        mired: VectorLike = MIRED_EXTENDED,
        sigfig: int = 5,
        planck_step: int = 1,
    ) -> None:
        """Initialize."""

        self.white = white
        self.table = self.generate_table(cmfs, white, mired, sigfig, planck_step)

    def generate_table(
        self,
        cmfs: dict[int, tuple[float, float, float]],
        white: VectorLike,
        mired: VectorLike,
        sigfig: int,
        planck_step: int,
    ) -> list[tuple[float, float, float, float]]:
        """
        Generate the necessary table for the Robertson1968 method.

        The below algorithm, coupled with the 1nm CMFs for the 1931 2 degree table, allows us to replicate the
        documented 31 points exactly.

        For each mired value we calculate two additional points, one on each side at a distance of 0.1.
        We use a very small distance so that we can approximate the slope. We calculate the distance between
        the targeted value and the two neighbors and then calculate the slope of the two small lines. Then we
        can calculate an interpolation factor and interpolate the slope for our target.

        We are able to calculate the uv pair for each mired point directly except for 0. 0 requires us to
        interpolate the values as it will cause a divide by zero in the Planckian locus. In this case, we
        assume a perfect 0.5 (middle) for our interpolation.
        """

        xyzw = util.xy_to_xyz(white)
        table = []  # type: list[tuple[float, float, float, float]]
        to_uv = util.xy_to_uv_1960 if self.CHROMATICITY == 'uv-1960' else util.xy_to_uv
        for t in mired:
            uv1 = to_uv(planck.temp_to_xy_planckian_locus(1e6 / (t - 0.01), cmfs, xyzw, step=planck_step))
            uv2 = to_uv(planck.temp_to_xy_planckian_locus(1e6 / (t + 0.01), cmfs, xyzw, step=planck_step))
            if t == 0:
                factor = 0.5
                uv = [alg.lerp(uv1[0], uv2[0], factor), alg.lerp(uv1[1], uv2[1], factor)]
            else:
                uv = to_uv(planck.temp_to_xy_planckian_locus(1e6 / t, cmfs, xyzw, step=planck_step))
                d1 = math.sqrt((uv[1] - uv1[1]) ** 2 + (uv[0] - uv1[0]) ** 2)
                d2 = math.sqrt((uv2[1] - uv[1]) ** 2 + (uv2[0] - uv[0]) ** 2)
                factor = d1 / (d1 + d2)
            m1 = -((uv[1] - uv1[1]) / (uv[0] - uv1[0])) ** -1
            m2 = -((uv2[1] - uv[1]) / (uv2[0] - uv[0])) ** -1
            m = alg.lerp(m1, m2, factor)
            if sigfig:
                template = '{{:.{}g}}'.format(sigfig)
                table.append(
                    (
                        float(t),
                        float(template.format(uv[0])),
                        float(template.format(uv[1])),
                        float(template.format(m))
                    )
                )
            else:
                table.append((t, uv[0], uv[1], m))
        return table

    def to_cct(self, color: Color, **kwargs: Any) -> Vector:
        """Calculate a color's CCT."""

        u, v = color.split_chromaticity(self.CHROMATICITY)[:-1]
        end = len(self.table) - 1
        slope_invert = False

        # Search for line pair coordinate is between.
        previous_di = temp = duv = 0.0

        for index, current in enumerate(self.table):
            # Get the distance
            # If a table was generated with values down to 1000K,
            # we would get a positive slope, so to keep logic the
            # same, adjust distance calculation such that negative
            # is still what we are looking for.
            if current[3] < 0:
                di = (v - current[2]) - current[3] * (u - current[1])
            else:
                slope_invert = True
                di = (current[2] - v) - current[3] * (current[1] - u)
            if index > 0 and (di <= 0.0 or index == end):
                # Calculate the required interpolation factor between the two lines
                previous = self.table[index - 1]
                current_denom = math.sqrt(1.0 + current[3] ** 2)
                di /= current_denom
                previous_denom = math.sqrt(1.0 + previous[3] ** 2)
                dip = previous_di / previous_denom
                factor = dip / (dip - di)

                # Calculate the temperature, if the mired value is zero
                # assume the maximum temperature of 100000K.
                mired = alg.lerp(previous[0], current[0], factor)
                temp = 1.0E6 / mired if mired > 0 else float('inf')

                # Interpolate the slope vectors
                dup = 1 / previous_denom
                dvp = previous[3] / previous_denom
                du = 1 / current_denom
                dv = current[3] / current_denom
                du = alg.lerp(dup, du, factor)
                dv = alg.lerp(dvp, dv, factor)
                denom = math.sqrt(du ** 2 + dv ** 2)
                du /= denom
                dv /= denom

                # Calculate Duv
                duv = (
                    du * (u - alg.lerp(previous[1], current[1], factor)) +
                    dv * (v - alg.lerp(previous[2], current[2], factor))
                )

                break

            # Save distance as previous
            previous_di = di

        return [temp, -duv if duv and not slope_invert else duv]

    def from_cct(
        self,
        color: type[Color],
        space: str,
        kelvin: float,
        duv: float,
        scale: bool,
        scale_space: str | None,
        **kwargs: Any
    ) -> Color:
        """Calculate a color that satisfies the CCT."""

        # Find inverse temperature to use as index.
        r = 1.0E6 / kelvin
        u = v = 0.0
        end = len(self.table) - 2

        for index, current in enumerate(self.table):
            future = self.table[index + 1]

            # Find the two isotherms that our target temp is between
            if r < future[0] or index == end:
                # Find relative weight between the two values
                f = (future[0] - r) / (future[0] - current[0])

                # Interpolate the uv coordinates of our target temperature
                u = alg.lerp(future[1], current[1], f)
                v = alg.lerp(future[2], current[2], f)

                # Calculate the offset along the slope
                if duv:
                    slope_invert = current[3] >= 0

                    # Calculate the slope vectors
                    u1 = 1.0
                    v1 = current[3]
                    length = math.sqrt(1.0 + v1 ** 2)
                    u1 /= length
                    v1 /= length

                    u2 = 1.0
                    v2 = future[3]
                    length = math.sqrt(1.0 + v2 ** 2)
                    u2 /= length
                    v2 /= length

                    # Find vector from the locus to our point.
                    du = alg.lerp(u2, u1, f)
                    dv = alg.lerp(v2, v1, f)
                    denom = math.sqrt(du ** 2 + dv ** 2)
                    du /= denom
                    dv /= denom

                    # Adjust the uv by the calculated offset
                    u += du * (-duv if not slope_invert else duv)
                    v += dv * (-duv if not slope_invert else duv)
                break

        return color.chromaticity(space, [u, v, 1], self.CHROMATICITY, scale=scale, scale_space=scale_space)



"""A98 RGB color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector


def lin_a98rgb(rgb: Vector) -> Vector:
    """Convert an array of a98-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return [alg.npow(val, 563 / 256) for val in rgb]


def gam_a98rgb(rgb: Vector) -> Vector:
    """Convert an array of linear-light a98-rgb  in the range 0.0-1.0 to gamma corrected form."""

    return [alg.npow(val, 256 / 563) for val in rgb]


class A98RGB(sRGB):
    """A98 RGB class."""

    BASE = "a98-rgb-linear"
    NAME = "a98-rgb"
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from A98 RGB."""

        return lin_a98rgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to A98 RGB."""

        return gam_a98rgb(coords)



"""Plot color space using Plotly."""
import sys
import argparse
from scipy.spatial import Delaunay
from plotly.figure_factory import create_trisurf as trisurf
import plotly.graph_objects as go
import math
import plotly.io as io
import os

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from coloraide.spaces import HSLish, HSVish, Cylindrical, Labish, LChish, RGBish  # noqa: E402
from coloraide import algebra as alg  # noqa: E402

FORCE_RECT = ('cmy',)


def create_custom_hsl(gamut):
    """Create a custom color object that has access to a special `hsl-gamut` space to map surface in."""

    cs = Color.CS_MAP[gamut]

    class HSL(type(Color.CS_MAP['hsl'])):
        NAME = 'hsl-{}'.format(gamut)
        BASE = gamut
        GAMUT_CHECK = gamut
        WHITE = cs.WHITE
        DYAMIC_RANGE = cs.DYNAMIC_RANGE

    class ColorCyl(Color):
        """Custom color."""

    ColorCyl.register(HSL())

    return ColorCyl


def get_face_color(cmap, simplex):
    """Get best color."""

    return Color.average([cmap[simplex[0]], cmap[simplex[1]], cmap[simplex[2]]], space='srgb').to_string(hex=True)


def cyl_disc(ColorCyl, space, gamut, location, resolution, opacity, edges):
    """
    Plot cylindrical disc on either top or bottom of an RGB cylinder.

    Expectation is either a HSL, HSV, or HSB style cylinder.
    """

    cs = ColorCyl.CS_MAP[space]
    if isinstance(cs, (HSVish, HSLish)):
        space_type = 'hslish/hsvish'
    else:
        space_type = space
    gspace = ColorCyl.CS_MAP[gamut]
    factor = gspace.channels[1].high

    zpos = 0.0 if location == 'bottom' else 1.0 * factor
    # HWB bottom disc will have a single point in the center that is a different colors at different hues. The mesh will
    # resolve one of them as the center, usually red. This will cause color averaging in the center of the disc to be
    # red-ish for all colors in the center. At lower resolutions, this is more noticeable. To avoid this, interpolate
    # rings very close to zero radius, but not zero radius. The mesh will still connect all the points near the center,
    # but it will be too small to see.
    start, end = 1.0 * factor, (1e-6 if space == 'hwb' and location == 'bottom' else 0.0)

    # Render the two halves of the disc
    hue_start, hue_end = 0, 360
    x = []
    y = []
    z = []
    u = []
    v = []
    cmap = []

    # Interpolate a circle on the outer edge
    c1 = ColorCyl(gamut, [hue_start, start, zpos])
    c2 = ColorCyl(gamut, [hue_end, start, zpos])
    chan_name = str(c1._space.channels[1])
    s1 = ColorCyl.steps([c1, c2], steps=resolution, space=gamut, hue='specified')
    s2 = [t.clone().set(chan_name, end) for t in s1]

    # Interpolate concentric circles to the center of the disc
    step = int(resolution / 2)
    for r in range(step):
        for t1, t2 in zip(s1, s2):
            c = t1.mix(t2, r / (step - 1), space=gamut, hue='specified')
            hue = c._space.hue_index()
            a, b = alg.polar_to_rect(c[1], c[hue])
            u.append(a)
            v.append(b)
            c.convert(space, norm=False, in_place=True)

            # HSL and HSV spaces
            if space_type == 'hslish/hsvish':
                hue, saturation, lightness = c._space.indexes()
                a, b = alg.polar_to_rect(c[saturation], c[hue])
                x.append(a)
                y.append(b)
                z.append(c[lightness])

            # HWB
            else:
                hue = c._space.hue_index()
                a, b = alg.polar_to_rect(c[1], c[hue])
                x.append(a)
                y.append(b)
                z.append(c[2])

            # Ensure colors fit in output color gamut.
            s = c.convert('srgb')
            if not s.in_gamut():
                s.fit()
            else:
                s.clip()
            cmap.append(s)

    # Calculate triangles
    tri = Delaunay(list(zip(u, v)))

    # Generate triangulated surface
    trace = trisurf(
        x=x, y=y, z=z,
        simplices=tri.simplices,
        show_colorbar=False,
        plot_edges=edges,
        color_func=[get_face_color(cmap, t) for t in tri.simplices]
    ).data
    trace[0].update(opacity=opacity)
    if edges:
        trace[1].update(hoverinfo='skip')

    return trace


def render_space_cyl(fig, space, gamut, resolution, opacity, edges):
    """
    Renders the color space using an RGB cylinder that is then mapped to the given space.

    Ideally used to represent cylindrical spaces and will align the the lightness equivalent
    as the Z axis. Lab-ish colors are performed in the mode as they are essentially cylindrical
    with the chroma and hue converted to Cartesian a and b.
    """

    target = Color.CS_MAP[space]
    is_cyl = isinstance(target, Cylindrical)
    is_labish = isinstance(target, Labish)
    is_lchish = isinstance(target, LChish)
    is_hslish = isinstance(target, HSLish)
    is_hslish_hsvish = is_hslish or isinstance(target, HSVish)

    # Determine the gamut mapping space to use.
    # Some spaces cannot be generalized (HWB and HPLuv for instance).
    if space in ('hwb', 'hpluv'):
        ColorCyl = Color
        gamut_space = space
    else:
        ColorCyl = create_custom_hsl(gamut)
        gamut_space = 'hsl-{}'.format(gamut)
    gspace = ColorCyl.CS_MAP[gamut_space]

    # Adjust scaling factor if the mapping space requires it
    factor = gspace.channels[1].high

    # Render the two halves of the cylinder
    start, end = 0, 360
    u = []
    v = []
    x = []
    y = []
    z = []
    cmap = []

    # Interpolate the cylinder from 0 to 360 degrees
    c1 = ColorCyl(gamut_space, [start, 1 * factor, 1 * factor])
    c2 = ColorCyl(gamut_space, [start, 1 * factor, 0])
    c3 = ColorCyl(gamut_space, [end, 1 * factor, 1 * factor])
    c4 = ColorCyl(gamut_space, [end, 1 * factor, 0])
    s1 = ColorCyl.steps([c1, c2], steps=resolution, space=gamut_space, hue='specified')
    s2 = ColorCyl.steps([c3, c4], steps=resolution, space=gamut_space, hue='specified')

    # Create a 3D mesh by interpolating ring at each lightness down the cylinder side.
    for t1, t2 in zip(s1, s2):
        for c in ColorCyl.steps([t1, t2], steps=resolution, space=gamut_space, hue='specified'):
            u.append(c[2])
            v.append(c['hue'])
            c.convert(space, norm=False, in_place=True)
            # LCh spaces
            if is_lchish:
                light, chroma, hue = c._space.names()
                a, b = alg.polar_to_rect(c[chroma], c[hue])
                x.append(a)
                y.append(b)
                z.append(c[light])

            # HSL and HSV spaces
            elif is_hslish_hsvish:
                hue, sat, light = c._space.names()
                a, b = alg.polar_to_rect(c[sat], c[hue])
                x.append(a)
                y.append(b)
                z.append(c[light])

            # HWB or any other cylindrical space that doesn't fit in the categories above.
            elif is_cyl:
                hue = c._space.hue_index()
                a, b = alg.polar_to_rect(c[1], c[hue])
                x.append(a)
                y.append(b)
                z.append(c[2])

            # Lab spaces
            elif is_labish:
                light, a, b = c._space.names()
                x.append(c[a])
                y.append(c[b])
                z.append(c[light])

            # Non-cylindrical spaces could be done here, but normally are not.
            else:
                x.append(c[0])
                y.append(c[1])
                z.append(c[2])

            # Adjust gamut to fit the display space
            s = c.convert('srgb')
            if not s.in_gamut():
                s.fit()
            else:
                s.clip()

            cmap.append(s)

    # Calculate the triangles
    tri = Delaunay(list(zip(u, v)))

    # Build the triangulated surface
    trace = trisurf(
        x=x, y=y, z=z,
        simplices=tri.simplices,
        show_colorbar=False,
        plot_edges=edges,
        color_func=[get_face_color(cmap, t) for t in tri.simplices]
    ).data
    trace[0].update(opacity=opacity)
    if edges:
        trace[1].update(hoverinfo='skip')
    fig.add_traces(trace)

    # Generate tops for spaces that do not normally get tops automatically.
    if space in ('hwb', 'hpluv'):
        fig.add_traces(cyl_disc(ColorCyl, space, gamut_space, 'top', resolution, opacity, edges))

    if is_cyl and not is_labish and not is_lchish:
        # We normally get a bottom except in the case of HWB.
        fig.add_traces(cyl_disc(ColorCyl, space, gamut_space, 'bottom', resolution, opacity, edges))

    return fig


def render_rect_face(s1, s2, dim, space, gamut, resolution, opacity, edges):
    """Render the RGB rectangular face."""

    x = []
    y = []
    z = []
    X = []
    Y = []
    Z = []
    cmap = []

    # Render an RGB face by taking two interpolated sides and interpolating the points across the face
    for c1, c2 in zip(s1, s2):
        for t in Color.steps([c1, c2], steps=int(resolution / 4), space=gamut):
            x.append(t[0])
            y.append(t[1])
            z.append(t[2])
            t.convert(space, norm=False, in_place=True)
            X.append(t[0])
            Y.append(t[1])
            Z.append(t[2])

            # Fit colors to output gamut
            s = t.convert('srgb')
            if not s.in_gamut():
                s.fit()
            else:
                s.clip()
            cmap.append(s)

    # Calculate triangles
    tri = Delaunay(list(zip(locals().get(dim[0]), locals().get(dim[1]))))

    # Generate triangulated surface
    trace = trisurf(
        x=X, y=Y, z=Z,
        simplices=tri.simplices,
        show_colorbar=False,
        plot_edges=edges,
        color_func=[get_face_color(cmap, t) for t in tri.simplices]
    ).data
    trace[0].update(opacity=opacity)
    if edges:
        trace[1].update(hoverinfo='skip')

    return trace


def render_space_rect(fig, space, gamut, resolution, opacity, edges):
    """Render rectangular space."""

    # Six corners of the RGB cube
    ck = Color(gamut, [0, 0, 0])
    cw = Color(gamut, [1, 1, 1])
    cr = Color(gamut, [1, 0, 0])
    cg = Color(gamut, [0, 1, 0])
    cb = Color(gamut, [0, 0, 1])
    cy = Color(gamut, [1, 1, 0])
    cc = Color(gamut, [0, 1, 1])
    cm = Color(gamut, [1, 0, 1])

    # Interpolate two sides of a given face and interpolate the rest
    s1 = Color.steps([cy, cw], steps=resolution, space=gamut)
    s2 = Color.steps([cg, cc], steps=resolution, space=gamut)
    s3 = Color.steps([cr, cm], steps=resolution, space=gamut)
    s4 = Color.steps([ck, cb], steps=resolution, space=gamut)
    fig.add_traces(render_rect_face(s1, s2, ('x', 'z'), space, gamut, resolution, opacity, edges))
    fig.add_traces(render_rect_face(s1, s3, ('y', 'z'), space, gamut, resolution, opacity, edges))
    fig.add_traces(render_rect_face(s3, s4, ('x', 'z'), space, gamut, resolution, opacity, edges))
    fig.add_traces(render_rect_face(s4, s2, ('y', 'z'), space, gamut, resolution, opacity, edges))
    s1 = Color.steps([cb, cc], steps=resolution, space=gamut)
    s2 = Color.steps([cm, cw], steps=resolution, space=gamut)
    fig.add_traces(render_rect_face(s1, s2, ('x', 'y'), space, gamut, resolution, opacity, edges))
    s1 = Color.steps([ck, cg], steps=resolution, space=gamut)
    s2 = Color.steps([cr, cy], steps=resolution, space=gamut)
    fig.add_traces(render_rect_face(s1, s2, ('x', 'y'), space, gamut, resolution, opacity, edges))

    return fig


def plot_gamut_in_space(
    space,
    gamut,
    title="",
    dark=False,
    resolution=200,
    opacity=1.0,
    edges=False,
    size=(800, 800),
    camera=None,
    aspect=None,
    projection='perspective'
):
    """Plot the given space in sRGB."""

    io.templates.default = 'plotly_dark' if dark else 'plotly'

    # I have no idea why this number causes HSL to lose its bottom
    if resolution == 50:
        resolution = 51

    if camera is None:
        camera = {'a': 45, 'e': 45, 'r': math.sqrt(1.25 ** 2 + 1.25 ** 2 + 1.25 ** 2)}

    a = math.radians((90 - camera['a']) % 360)
    e = math.radians(90 - camera['e'])
    r = camera['r']
    y = r * math.sin(e) * math.cos(a)
    x = r * math.sin(e) * math.sin(a)
    z = r * math.cos(e)

    if aspect is None:
        aspect = {'x': 1, 'y': 1, 'z': 1}

    # Get names for
    target = Color.CS_MAP[space]
    if len(target.CHANNELS) > 3:
        print('Color spaces with dimensions greater than 3 are not supported')
        return None

    names = target.CHANNELS
    is_rgbish = isinstance(target, RGBish)
    is_cyl = isinstance(target, Cylindrical)
    is_labish = isinstance(target, Labish)
    is_lchish = isinstance(target, LChish)
    is_hslish_hsvish = isinstance(target, (HSLish, HSVish))

    # Setup axes
    if is_labish:
        c1, c2, c3 = target.indexes()
        axm = [c2, c3, c1]
    elif is_lchish:
        c1, c2, c3 = target.indexes()
        axm = [c3, c2, c1]
    elif is_hslish_hsvish:
        axm = [0, 1, 2]
    else:
        axm = [0, 1, 2]

    showbackground = True
    backgroundcolor = "rgb(230, 230, 230)" if not dark else '#282830'
    gridcolor = "rgb(255, 255, 255)" if not dark else '#111'
    zerolinecolor = "rgb(255, 255, 255)" if not dark else '#111'
    axis = {
        "showbackground": showbackground,
        "backgroundcolor": backgroundcolor,
        "gridcolor": gridcolor,
        "zerolinecolor": zerolinecolor,
    }
    xaxis = str(names[axm[0]]) if not is_cyl else "{} (0˚ - 360˚)".format(names[axm[0]])
    yaxis = str(names[axm[1]])
    zaxis = str(names[axm[2]])

    # Setup plot layout
    layout = go.Layout(
        # General figure characteristics
        title=title,
        width=size[0],
        height=size[1],

        # Specify scene layout
        scene=go.layout.Scene(
            xaxis=go.layout.scene.XAxis(title=xaxis, showticklabels=not is_cyl, **axis),
            yaxis=go.layout.scene.YAxis(title=yaxis, **axis),
            zaxis=go.layout.scene.ZAxis(title=zaxis, **axis),
            aspectratio=aspect
        ),

        # Control camera position
        scene_camera={
            "projection": go.layout.scene.camera.Projection(type=projection),
            "center": {"x": 0, "y": 0, "z": 0},
            "up": {"x": 0, "y": 0, "z": 1},
            "eye": {"x": x, "y": y, "z": z}
        }
    )

    # Create figure to store the plot
    fig = go.Figure(layout=layout)

    target = Color.CS_MAP[space]
    if is_rgbish or space in FORCE_RECT:
        # Use a rectangular space for RGB-ish spaces to give a sharper cube
        return render_space_rect(fig, space, gamut, resolution, opacity, edges)
    else:
        # Render the space plot using a cylindrical space as the gamut space
        return render_space_cyl(fig, space, gamut, resolution, opacity, edges)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='3d_diagrams', description='Plot 3D gamut in a different color spaces.')
    parser.add_argument('--space', '-s', help='Desired space.')
    parser.add_argument(
        '--gamut',
        '-g',
        default='srgb',
        help=(
            'Gamut space to render space in. Gamut space must be bounded and must have channels in the range [0, 1].'
            'As only a shell is rendered for the gamut, the target space should be less than or equal to the size of '
            'the target gamut or there will be areas that do not render. Cylindrical spaces based specifically off '
            'an RGB gamut, such as HSL being based on sRGB, will only be done under the related gamut and will ignore '
            'this option.'
        )
    )
    parser.add_argument('--opacity', default=1.0, type=float, help="opacity")
    parser.add_argument(
        '--resolution', '-r',
        default="200",
        help=(
            "How densely to render the figure. Some spaces need higher resolution to flesh out certain areas, "
            "but it comes at the cost of speed. Minimum is 60, default is 200."
        )
    )
    parser.add_argument('--pos', '-p', default=None, help="Position of camara 'x:y:z'")
    parser.add_argument('--edges', '-e', action="store_true", help="Plot edges.")
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--output', '-o', default='', help='Output file.')
    parser.add_argument('--height', '-H', type=int, default=800, help="Height")
    parser.add_argument('--width', '-W', type=int, default=800, help="Width")
    parser.add_argument('--azimuth', '-A', type=float, default=45, help="Camera X position")
    parser.add_argument('--elevation', '-E', type=float, default=45, help="Camera Y position")
    parser.add_argument('--distance', '-D', type=float, default=2.5, help="Camera Z position")
    parser.add_argument(
        '--aspect-ratio', '-R',
        default='1:1:1',
        help="Aspect ratio. Set to 0:0:0 to leave aspect ratio untouched."
    )
    parser.add_argument(
        '--projection', '-P',
        default='perspective',
        help="Projection mode, perspective or orthographic"
    )
    args = parser.parse_args()

    aspect = {k: float(v) for k, v in zip(['x', 'y', 'z'], args.aspect_ratio.split(':'))}

    # Plot the color space
    fig = plot_gamut_in_space(
        args.space,
        args.gamut,
        title=args.title,
        dark=args.dark,
        resolution=max(8, int(args.resolution)),
        opacity=args.opacity,
        edges=args.edges,
        size=(args.width, args.height),
        camera={'a': args.azimuth, 'e': args.elevation, 'r': args.distance},
        aspect=aspect,
        projection=args.projection
    )

    # Show or save the data as an image, etc.
    if fig:
        if args.output:
            filetype = os.path.splitext(args.output)[1].lstrip('.').lower()
            if filetype == 'html':
                with open(args.output, 'w') as f:
                    f.write(io.to_html(fig))
            elif filetype == 'json':
                io.write_json(fig, args.output)
            else:
                with open(args.output, 'wb') as f:
                    f.write(fig.to_image(format=filetype))
        else:
            fig.show()
        return 0
    return 1


if __name__ == "__main__":
    sys.exit(main())



"""Test Prismatic."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestPrismatic(util.ColorAssertsPyTest):
    """Test Prismatic."""

    COLORS = [
        ('red', 'color(--prismatic 1 1 0 0)'),
        ('orange', 'color(--prismatic 1 0.60714 0.39286 0)'),
        ('yellow', 'color(--prismatic 1 0.5 0.5 0)'),
        ('green', 'color(--prismatic 0.50196 0 1 0)'),
        ('blue', 'color(--prismatic 1 0 0 1)'),
        ('indigo', 'color(--prismatic 0.5098 0.36585 0 0.63415)'),
        ('violet', 'color(--prismatic 0.93333 0.39274 0.21452 0.39274)'),
        ('white', 'color(--prismatic 1 0.33333 0.33333 0.33333)'),
        ('gray', 'color(--prismatic 0.50196 0.33333 0.33333 0.33333)'),
        ('black', 'color(--prismatic 0 0 0 0)'),
        # Test CSS color
        ('color(--prismatic 0 0.50196 0 0)', 'color(--prismatic 0 0.50196 0 0)'),
        ('color(--prismatic 0 0.50196 0 0 / 0.5)', 'color(--prismatic 0 0.50196 0 0 / 0.5)'),
        ('color(--prismatic 50% 50% 50% 50% / 50%)', 'color(--prismatic 0.5 0.5 0.5 0.5 / 0.5)'),
        ('color(--prismatic none none none none / none)', 'color(--prismatic none none none none / none)'),
        # Test range
        ('color(--prismatic 0% 0% 0% 0%)', 'color(--prismatic 0 0 0 0)'),
        ('color(--prismatic 100% 100% 100% 100%)', 'color(--prismatic 1 1 1 1)'),
        ('color(--prismatic -100% -100% -100% -100%)', 'color(--prismatic -1 -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_prismatic_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('prismatic'), Color(color2))


class TestPrismaticSerialize(util.ColorAssertsPyTest):
    """Test Prismatic serialization."""

    COLORS = [
        # Test color
        ('color(--prismatic 0 0.3 0.75 0.5 / 0.5)', {}, 'color(--prismatic 0 0.3 0.75 0.5 / 0.5)'),
        # Test alpha
        ('color(--prismatic 0 0.3 0.75 0.5)', {'alpha': True}, 'color(--prismatic 0 0.3 0.75 0.5 / 1)'),
        ('color(--prismatic 0 0.3 0.75 0.5 / 0.5)', {'alpha': False}, 'color(--prismatic 0 0.3 0.75 0.5)'),
        # Test None
        ('color(--prismatic none 0.3 0.75 0.5)', {}, 'color(--prismatic 0 0.3 0.75 0.5)'),
        ('color(--prismatic none 0.3 0.75 0.5)', {'none': True}, 'color(--prismatic none 0.3 0.75 0.5)'),
        # Test Fit
        ('color(--prismatic 1.2 0.2 0 0.5)', {}, 'color(--prismatic 1 0.29182 0.11319 0.595)'),
        ('color(--prismatic 1.2 0.2 0 0.5)', {'fit': False}, 'color(--prismatic 1.2 0.2 0 0.5)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestPrismaticPoperties(util.ColorAsserts, unittest.TestCase):
    """Test Prismatic."""

    def test_l(self):
        """Test `l`."""

        c = Color('color(--prismatic 1 0.5 0.5 0)')
        self.assertEqual(c['l'], 1)
        c['l'] = 0.2
        self.assertEqual(c['l'], 0.2)

    def test_r(self):
        """Test `r`."""

        c = Color('color(--prismatic 1 0.5 0.5 0)')
        self.assertEqual(c['r'], 0.5)
        c['r'] = 0.1
        self.assertEqual(c['r'], 0.1)

    def test_g(self):
        """Test `g`."""

        c = Color('color(--prismatic 1 0.5 0.5 0)')
        self.assertEqual(c['g'], 0.5)
        c['g'] = 0.1
        self.assertEqual(c['g'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--prismatic 1 0.5 0.5 0)')
        self.assertEqual(c['b'], 0)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--prismatic 1 0.5 0.5 0)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('prismatic', [0.3, 0.3, 0.3, 0.3]).is_achromatic(), True)
        self.assertEqual(Color('prismatic', [0.3, 0.3000001, 0.3, 0.3]).is_achromatic(), True)
        self.assertEqual(Color('prismatic', [NaN, 0.3, 0.3, 0.3]).is_achromatic(), True)
        self.assertEqual(Color('prismatic', [0.0, 0.4, 0.3, 0.3]).is_achromatic(), True)
        self.assertEqual(Color('prismatic', [0.0, NaN, 0.3, 0.0]).is_achromatic(), True)
        self.assertEqual(Color('prismatic', [0.3, 0.4, 0.3, 0.3]).is_achromatic(), False)
        self.assertEqual(Color('prismatic', [0.3, NaN, 0.3, 0.3]).is_achromatic(), False)



"""XYZ class."""
from __future__ import annotations
from ..cat import WHITES
from .xyz_d65 import XYZD65


class XYZD50(XYZD65):
    """XYZ D50 class."""

    BASE = "xyz-d65"
    NAME = "xyz-d50"
    SERIALIZE = ("xyz-d50",)
    WHITE = WHITES['2deg']['D50']



"""
Hunter Lab class.

https://support.hunterlab.com/hc/en-us/articles/203997095-Hunter-Lab-Color-Scale-an08-96a2
"""
from __future__ import annotations
from ..cat import WHITES
from ..spaces.lab import Lab
from .. import algebra as alg
from .. import util
from ..types import Vector, VectorLike
from ..channels import Channel, FLG_MIRROR_PERCENT

# Values for the original Hunter Lab with illuminant C.
# Used to calculate an appropriate `Ka` and `Kb` for whatever white point we are using.
CXN = 98.04
CYN = 100.0
CZN = 118.11
CKA = 175.0
CKB = 70.0


def xyz_to_hlab(xyz: Vector, white: VectorLike) -> Vector:
    """Convert XYZ to Hunter Lab."""

    xn, yn, zn = alg.multiply(util.xy_to_xyz(white), 100, dims=alg.D1_SC)
    ka = CKA * alg.nth_root(xn / CXN, 2)
    kb = CKB * alg.nth_root(zn / CZN, 2)
    x, y, z = alg.multiply(xyz, 100, dims=alg.D1_SC)
    l = alg.nth_root(y / yn, 2)
    a = b = 0.0
    if l != 0:
        a = ka * (x / xn - y / yn) / l
        b = kb * (y / yn - z / zn) / l
    return [l * 100, a, b]


def hlab_to_xyz(hlab: Vector, white: VectorLike) -> Vector:
    """Convert Hunter Lab to XYZ."""

    xn, yn, zn = alg.multiply(util.xy_to_xyz(white), 100, dims=alg.D1_SC)
    ka = CKA * alg.nth_root(xn / CXN, 2)
    kb = CKB * alg.nth_root(zn / CZN, 2)
    l, a, b = hlab
    l /= 100
    y = (l ** 2) * yn
    x = (((a * l) / ka) + (y / yn)) * xn
    z = (((b * l) / kb) - (y / yn)) * -zn
    return alg.divide([x, y, z], 100, dims=alg.D1_SC)


class HunterLab(Lab):
    """Hunter Lab class."""

    BASE = 'xyz-d65'
    NAME = "hunter-lab"
    SERIALIZE = ("--hunter-lab",)
    WHITE = WHITES['2deg']['D65']
    CHANNELS = (
        Channel("l", 0.0, 100.0),
        Channel("a", -210.0, 210.0, flags=FLG_MIRROR_PERCENT),
        Channel("b", -210.0, 210.0, flags=FLG_MIRROR_PERCENT)
    )

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Hunter Lab."""

        return hlab_to_xyz(coords, self.white())

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Hunter Lab."""

        return xyz_to_hlab(coords, self.white())



"""Lab class."""
from __future__ import annotations
from ...spaces import Space, Labish
from ...cat import WHITES
from ...channels import Channel, FLG_OPT_PERCENT, FLG_MIRROR_PERCENT
from ... import util
from ... import algebra as alg
from ...types import VectorLike, Vector

ACHROMATIC_THRESHOLD = 1e-4
EPSILON = 216 / 24389  # `6^3 / 29^3`
EPSILON3 = 6 / 29  # Cube root of EPSILON
KAPPA = 24389 / 27
KE = 8  # KAPPA * EPSILON = 8


def lab_to_xyz(lab: Vector, white: VectorLike) -> Vector:
    """
    Convert Lab to D50-adapted XYZ.

    http://www.brucelindbloom.com/Eqn_Lab_to_XYZ.html

    While the derivation is different than the specification, the results are the same as Appendix D:
    https://www.cdvplus.cz/file/3-publikace-cie15-2004/
    """

    l, a, b = lab

    # compute `f`, starting with the luminance-related term
    fy = (l + 16) / 116
    fx = a / 500 + fy
    fz = fy - b / 200

    # compute `xyz`
    xyz = [
        fx ** 3 if fx > EPSILON3 else (116 * fx - 16) / KAPPA,
        fy ** 3 if l > KE else l / KAPPA,
        fz ** 3 if fz > EPSILON3 else (116 * fz - 16) / KAPPA
    ]

    # Compute XYZ by scaling `xyz` by reference `white`
    return alg.multiply(xyz, white, dims=alg.D1)


def xyz_to_lab(xyz: Vector, white: VectorLike) -> Vector:
    """
    Assuming XYZ is relative to D50, convert to CIELab from CIE standard.

    http://www.brucelindbloom.com/Eqn_XYZ_to_Lab.html

    While the derivation is different than the specification, the results are the same:
    https://www.cdvplus.cz/file/3-publikace-cie15-2004/
    """

    # compute `xyz`, which is XYZ scaled relative to reference white
    xyz = alg.divide(xyz, white, dims=alg.D1)
    # Compute `fx`, `fy`, and `fz`
    fx, fy, fz = [alg.nth_root(i, 3) if i > EPSILON else (KAPPA * i + 16) / 116 for i in xyz]

    return [
        (116.0 * fy) - 16.0,
        500.0 * (fx - fy),
        200.0 * (fy - fz)
    ]


class Lab(Labish, Space):
    """Lab class."""

    BASE = "xyz-d50"
    NAME = "lab"
    SERIALIZE = ("--lab",)
    CHANNELS = (
        Channel("l", 0.0, 100.0, flags=FLG_OPT_PERCENT),
        Channel("a", -125.0, 125.0, flags=FLG_MIRROR_PERCENT | FLG_OPT_PERCENT),
        Channel("b", -125.0, 125.0, flags=FLG_MIRROR_PERCENT | FLG_OPT_PERCENT)
    )
    CHANNEL_ALIASES = {
        "lightness": "l"
    }
    WHITE = WHITES['2deg']['D50']

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return alg.rect_to_polar(coords[1], coords[2])[0] < ACHROMATIC_THRESHOLD

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ D50 from Lab."""

        return lab_to_xyz(coords, util.xy_to_xyz(self.white()))

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ D50 to Lab."""

        return xyz_to_lab(coords, util.xy_to_xyz(self.white()))



"""
Rec. 2100 PQ color class.

https://www.itu.int/dms_pubrec/itu-r/rec/bt/R-REC-BT.2100-2-201807-I!!PDF-E.pdf
"""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector
from .. import util


class Rec2100PQ(sRGB):
    """Rec. 2100 PQ class."""

    BASE = "rec2020-linear"
    NAME = "rec2100-pq"
    SERIALIZE = ('--rec2100-pq',)
    WHITE = WHITES['2deg']['D65']
    DYNAMIC_RANGE = 'hdr'

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Rec. 2100 PQ."""

        return alg.divide(util.pq_st2084_eotf(coords), util.YW, dims=alg.D1_SC)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Rec. 2100 PQ."""

        return util.pq_st2084_oetf(alg.multiply(coords, util.YW, dims=alg.D1_SC))



"""
Model interpolation.

Currently only shows non polar interpolation.
"""
import sys
import os
import matplotlib.pyplot as plt
import argparse
import math

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from tools.slice_diagram import plot_slice  # noqa: E402
from coloraide.spaces import Cylindrical  # noqa: E402
from coloraide import algebra as alg  # noqa: E402


def get_spline(x, y, steps=100):
    """Get spline."""

    return tuple(list(i) for i in zip(*alg.interpolate(list(zip(x, y)), method='monotone').steps(steps)))


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='gamut_diagrams', description='Demonstrate gamut mapping.')
    parser.add_argument('--space', '-s', default='srgb', help="Space to interpolate in.")
    parser.add_argument('--color', '-c', action='append', help="Color.")
    parser.add_argument('--position', '-p', type=float, help="Position between to show interpolated.")
    parser.add_argument('--gamut', '-g', default="srgb", help='Gamut to evaluate the color in (default is sRGB).')
    parser.add_argument('--method', '-m', default='linear', help="Interplation method to use: linear, bezier, etc.")
    parser.add_argument('--extrapolate', '-e', action='store_true', help='Extrapolate values.')
    parser.add_argument('--title', '-T', default='', help="Provide a title for the diagram.")
    parser.add_argument('--subtitle', '-t', default='', help="Provide a subtitle for the diagram.")
    parser.add_argument(
        '--constant',
        '-k',
        help=(
            "The channel to hold constant and the value to use 'name:value'. This will overwrite whatever the start"
            " and end color specifies as the plot must fit in the 2D plane."
        )
    )
    parser.add_argument('--xaxis', '-x', help="The channel to plot on X axis 'name:min:max'.")
    parser.add_argument('--yaxis', '-y', help="The channel to plot on Y axis 'name:min:max'.")
    parser.add_argument('--resolution', '-r', default="800", help="How densely to render the figure.")
    parser.add_argument('--no-border', '-b', action="store_true", help='Draw no border around the graphed content.')
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()

    colors = []
    for color in args.color:
        c = Color(color).convert(args.space)
        colors.append(c)

    for chan in args.constant.split(';'):
        c_name, c_value = chan.split(':')
        c_value = float(c_value)
        for color in colors:
            color.set(c_name, c_value)

    if not args.title:
        title = "Interpolation in the {} Color Space Using the {} Gamut".format(args.space, args.gamut)
    else:
        title = args.title

    plot_slice(
        args.space,
        args.constant,
        args.xaxis,
        args.yaxis,
        gamut=args.gamut,
        resolution=int(args.resolution),
        title=title,
        subtitle=args.subtitle,
        dark=args.dark,
        polar=True,
        border=not args.no_border
    )

    # Get the actual indexes of the specified channels
    c1 = colors[0]
    name1 = args.xaxis.split(":", 1)[0]
    name2 = args.yaxis.split(":", 1)[0]
    name1 = c1._space.CHANNEL_ALIASES.get(name1, name1)
    index1 = c1._space.get_channel_index(name1)
    name2 = c1._space.CHANNEL_ALIASES.get(name2, name2)
    hue_index = -1

    if isinstance(c1._space, Cylindrical):
        index = c1._space.hue_index()
        if index == index1:
            hue_index = index

    xs = []
    ys = []
    i = Color.interpolate(colors, space=args.space, method=args.method, extrapolate=args.extrapolate)
    if not args.extrapolate:
        offset, factor = 0, 1
    else:
        offset, factor = 1, 3
    for r in range(101):
        c = i((r * factor / 100) - offset)
        xs.append(c.get(name1) if hue_index == -1 else math.radians(c.get(name1)))
        ys.append(c.get(name2))

    xs, ys = get_spline(xs, ys, len(xs) * 3)

    plt.plot(
        xs,
        ys,
        color='black',
        marker="",
        linewidth=1.5,
        markersize=2,
        antialiased=True
    )

    for c in colors:
        plt.scatter(
            c.get(name1) if hue_index == -1 else math.radians(c.get(name1)),
            c.get(name2),
            marker="o",
            color=c.convert('srgb').to_string(hex=True),
            edgecolor='black',
            zorder=100
        )

    if args.position is not None:
        cp = Color.interpolate(colors, space=args.space, method=args.method)(float(args.position))
        plt.scatter(
            cp.get(name1) if hue_index == -1 else math.radians(cp.get(name1)),
            cp.get(name2),
            marker="o",
            color=cp.convert('srgb').to_string(hex=True),
            edgecolor='black',
            s=64,
            zorder=100
        )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Test oRGB."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
import pytest


class TestoRGB(util.ColorAssertsPyTest):
    """Test oRGB."""

    COLORS = [
        ('red', 'color(--orgb 0.299 0.00002 0.99998)'),
        ('orange', 'color(--orgb 0.67882 0.75654 0.4464)'),
        ('yellow', 'color(--orgb 0.886 1 0)'),
        ('green', 'color(--orgb 0.29465 0.00001 -0.50195)'),
        ('blue', 'color(--orgb 0.114 -1 0)'),
        ('indigo', 'color(--orgb 0.14606 -0.39733 0.19643)'),
        ('violet', 'color(--orgb 0.68472 -0.29948 0.29947)'),
        ('white', 'color(--orgb 1 0 0)'),
        ('gray', 'color(--orgb 0.50196 0 0)'),
        ('black', 'color(--orgb 0 0 0)'),
        # Test color
        ('color(--orgb 0.5 0.1 -0.1)', 'color(--orgb 0.5 0.1 -0.1)'),
        ('color(--orgb 0.5 0.1 -0.1 / 0.5)', 'color(--orgb 0.5 0.1 -0.1 / 0.5)'),
        ('color(--orgb 50% 50% -50% / 50%)', 'color(--orgb 0.5 0.5 -0.5 / 0.5)'),
        ('color(--orgb none none none / none)', 'color(--orgb none none none / none)'),
        # Test percent ranges
        ('color(--orgb 0% 0% 0%)', 'color(--orgb 0 0 0)'),
        ('color(--orgb 100% 100% 100%)', 'color(--orgb 1 1 1)'),
        ('color(--orgb -100% -100% -100%)', 'color(--orgb -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_orgb_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('orgb'), Color(color2))


class TestoRGBSerialize(util.ColorAssertsPyTest):
    """Test oRGB serialization."""

    COLORS = [
        # Test color
        ('color(--orgb 0.75 0.1 -0.1 / 0.5)', {}, 'color(--orgb 0.75 0.1 -0.1 / 0.5)'),
        # Test alpha
        ('color(--orgb 0.75 0.1 -0.1)', {'alpha': True}, 'color(--orgb 0.75 0.1 -0.1 / 1)'),
        ('color(--orgb 0.75 0.1 -0.1 / 0.5)', {'alpha': False}, 'color(--orgb 0.75 0.1 -0.1)'),
        # Test None
        ('color(--orgb none 0.1 -0.1)', {}, 'color(--orgb 0 0.1 -0.1)'),
        ('color(--orgb none 0.1 -0.1)', {'none': True}, 'color(--orgb none 0.1 -0.1)'),
        # Test Fit
        ('color(--orgb 0.75 1.2 -0.1)', {}, 'color(--orgb 0.76398 1 -0.05221)'),
        ('color(--orgb 0.75 1.2 -0.1)', {'fit': False}, 'color(--orgb 0.75 1.2 -0.1)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestoRGBPoperties(util.ColorAsserts, unittest.TestCase):
    """Test oRGB."""

    def test_l(self):
        """Test `l`."""

        c = Color('color(--orgb 1 0.5 0.5)')
        self.assertEqual(c['l'], 1)
        c['l'] = 0.2
        self.assertEqual(c['l'], 0.2)

    def test_cyb(self):
        """Test `cyb`."""

        c = Color('color(--orgb 1 0.5 0.5)')
        self.assertEqual(c['cyb'], 0.5)
        c['cyb'] = 0.1
        self.assertEqual(c['cyb'], 0.1)

    def test_crg(self):
        """Test `crg`."""

        c = Color('color(--orgb 1 0.5 0.5)')
        self.assertEqual(c['crg'], 0.5)
        c['crg'] = 0.1
        self.assertEqual(c['crg'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--orgb 1 0.5 0.5)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)

    def test_labish_names(self):
        """Test `labish_names`."""

        c = Color('color(--orgb 1 0.5 0.5)')
        self.assertEqual(c._space.names(), ('l', 'crg', 'cyb'))



"""Test LChuv library."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
from coloraide import NaN
import pytest


class TestLChuv(util.ColorAssertsPyTest):
    """Test LChuv."""

    COLORS = [
        ('red', 'color(--lchuv 53.237 179.04 12.177)'),
        ('orange', 'color(--lchuv 74.934 105.26 44.683)'),
        ('yellow', 'color(--lchuv 97.139 107.09 85.874)'),
        ('green', 'color(--lchuv 46.228 71.548 127.72)'),
        ('blue', 'color(--lchuv 32.301 130.69 265.87)'),
        ('indigo', 'color(--lchuv 20.47 62.163 279.33)'),
        ('violet', 'color(--lchuv 69.695 84.743 307.72)'),
        ('white', 'color(--lchuv 100 0 none)'),
        ('gray', 'color(--lchuv 53.585 0 none)'),
        ('black', 'color(--lchuv 0 0 none)'),
        # Test color
        ('color(--lchuv 100 30 270)', 'color(--lchuv 100 30 270)'),
        ('color(--lchuv 100 30 270 / 0.5)', 'color(--lchuv 100 30 270 / 0.5)'),
        ('color(--lchuv 50% 50% 50% / 50%)', 'color(--lchuv 50 110 180 / 0.5)'),
        ('color(--lchuv none none none / none)', 'color(--lchuv none none none / none)'),
        # Test percent ranges
        ('color(--lchuv 0% 0% 0%)', 'color(--lchuv 0 0 none)'),
        ('color(--lchuv 100% 100% 100%)', 'color(--lchuv 100 220 360 / 1)'),
        ('color(--lchuv -100% -100% -100%)', 'color(--lchuv -100 0 -360 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('lchuv'), Color(color2))


class TestLChuvSerialize(util.ColorAssertsPyTest):
    """Test LChuv serialization."""

    COLORS = [
        # Test color
        ('color(--lchuv 75 30 50 / 0.5)', {}, 'color(--lchuv 75 30 50 / 0.5)'),
        # Test alpha
        ('color(--lchuv 75 30 50)', {'alpha': True}, 'color(--lchuv 75 30 50 / 1)'),
        ('color(--lchuv 75 30 50 / 0.5)', {'alpha': False}, 'color(--lchuv 75 30 50)'),
        # Test None
        ('color(--lchuv none 30 50)', {}, 'color(--lchuv 0 30 50)'),
        ('color(--lchuv none 30 50)', {'none': True}, 'color(--lchuv none 30 50)'),
        # Test Fit (not bound)
        ('color(--lchuv 75 230 50)', {}, 'color(--lchuv 75 230 50)'),
        ('color(--lchuv 75 230 50)', {'fit': False}, 'color(--lchuv 75 230 50)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestLChuvD65Properties(util.ColorAsserts, unittest.TestCase):
    """Test LChuv."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--lchuv 90% 50 120 / 1)')
        self.assertEqual(c['lightness'], 90)
        c['lightness'] = 80
        self.assertEqual(c['lightness'], 80)

    def test_chroma(self):
        """Test `chroma`."""

        c = Color('color(--lchuv 90% 50 120 / 1)')
        self.assertEqual(c['chroma'], 50)
        c['chroma'] = 40
        self.assertEqual(c['chroma'], 40)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--lchuv 90% 50 120 / 1)')
        self.assertEqual(c['hue'], 120)
        c['hue'] = 110
        self.assertEqual(c['hue'], 110)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--lchuv 90% 50 120 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('lchuv', [90, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--lchuv 90% 0 none / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_near_zero_null(self):
        """
        Test very near zero null.

        This is a fix up to help give more sane hues
        when chroma is very close to zero.
        """

        c = Color('color(--lchuv 90% 0.000000000009 120 / 1)').convert('luv').convert('lchuv')
        self.assertTrue(c.is_nan('hue'))

    def test_from_luv(self):
        """Test null from Luv conversion."""

        c1 = Color('color(--luv 90% 0 0)')
        c2 = c1.convert('lchuv')
        self.assertColorEqual(c2, Color('color(--lchuv 90% 0 0)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_null_normalization_min_chroma(self):
        """Test minimum saturation."""

        c = Color('color(--lchuv 90% 0 120 / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to LChuv with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('lchuv')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to lab."""

        self.assertColorEqual(Color('color(--lchuv 90% -10 120 / 1)').convert('luv'), Color('color(--luv 90% 0 0)'))


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('lchuv', [30, 0, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [30, 0.000001, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [NaN, 0.00001, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [0, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [0, 100, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [NaN, 20, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [30, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('lchuv', [NaN, NaN, 270]).is_achromatic(), True)



"""HSL class."""
from __future__ import annotations
from ...spaces import HSLish, Space
from ...cat import WHITES
from ...channels import Channel, FLG_ANGLE, FLG_OPT_PERCENT
from ... import util
from ...types import Vector


def srgb_to_hsl(rgb: Vector) -> Vector:
    """Convert sRGB to HSL."""

    r, g, b = rgb
    mx = max(rgb)
    mn = min(rgb)
    h = 0.0
    s = 0.0
    l = (mn + mx) / 2
    c = mx - mn

    if c != 0.0:
        if mx == r:
            h = (g - b) / c
        elif mx == g:
            h = (b - r) / c + 2.0
        else:
            h = (r - g) / c + 4.0
        s = 0 if l == 0.0 or l == 1.0 else (mx - l) / min(l, 1 - l)
        h *= 60.0

    return [util.constrain_hue(h), s, l]


def hsl_to_srgb(hsl: Vector) -> Vector:
    """
    HSL to RGB.

    https://en.wikipedia.org/wiki/HSL_and_HSV#HSL_to_RGB_alternative
    """

    h, s, l = hsl
    h = util.constrain_hue(h)

    def f(n: int) -> float:
        """Calculate the channels."""
        k = (n + h / 30) % 12
        a = s * min(l, 1 - l)
        return l - a * max(-1, min(k - 3, 9 - k, 1))

    return [f(0), f(8), f(4)]


class HSL(HSLish, Space):
    """HSL class."""

    BASE = "srgb"
    NAME = "hsl"
    SERIALIZE = ("--hsl",)
    CHANNELS = (
        Channel("h", 0.0, 360.0, bound=True, flags=FLG_ANGLE),
        Channel("s", 0.0, 1.0, bound=True, flags=FLG_OPT_PERCENT),
        Channel("l", 0.0, 1.0, bound=True, flags=FLG_OPT_PERCENT)
    )
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "lightness": "l"
    }
    WHITE = WHITES['2deg']['D65']
    GAMUT_CHECK = "srgb"

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return abs(coords[1]) < 1e-4 or coords[2] == 0.0 or abs(1 - coords[2]) < 1e-7

    def to_base(self, coords: Vector) -> Vector:
        """To sRGB from HSL."""

        return hsl_to_srgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From sRGB to HSL."""

        return srgb_to_hsl(coords)



"""ColorAide Library."""
from .__meta__ import __version_info__, __version__  # noqa: F401
from .color import Color, ColorMatch
from .interpolate import stop, hint
from .algebra import NaN
from .easing import cubic_bezier, linear, ease, ease_in, ease_out, ease_in_out

__all__ = (
    "Color", "ColorMatch", "NaN", "stop", "hint", "cubic_bezier",
    "linear", "ease", "ease_in", "ease_out", "ease_in_out"
)



"""
ACEScct color space.

https://www.oscars.org/science-technology/aces/aces-documentation
"""
from __future__ import annotations
import math
from ..channels import Channel
from ..spaces.srgb import sRGB
from ..types import Vector
from .acescc import CC_MAX

CCT_MIN = 0.0729055341958355
CCT_MAX = CC_MAX
C1 = 0.0078125
C2 = 10.5402377416545
C3 = 0.155251141552511


def acescct_to_acescg(acescc: Vector) -> Vector:
    """Convert ACEScc to ACEScg."""

    acescg = []
    for c in acescc:
        if c <= C3:
            c = (c - CCT_MIN) / C2
        elif C3 <= c < CCT_MAX:
            c = 2 ** (c * 17.52 - 9.72)
        else:
            c = 65504
        acescg.append(c)
    return acescg


def acescg_to_acescct(acescg: Vector) -> Vector:
    """Convert ACEScg to ACEScc."""

    acescc = []
    for c in acescg:
        if c <= C1:
            c = C2 * c + CCT_MIN
        elif c > C1:
            c = (math.log2(c) + 9.72) / 17.52
        acescc.append(c)
    return acescc


class ACEScct(sRGB):
    """The ACEScct color class."""

    BASE = "acescg"
    NAME = "acescct"
    SERIALIZE = ("--acescct",)  # type: tuple[str, ...]
    WHITE = (0.32168, 0.33767)
    CHANNELS = (
        Channel("r", CCT_MIN, CCT_MAX, bound=True, nans=CCT_MIN),
        Channel("g", CCT_MIN, CCT_MAX, bound=True, nans=CCT_MIN),
        Channel("b", CCT_MIN, CCT_MAX, bound=True, nans=CCT_MIN)
    )
    DYNAMIC_RANGE = 'hdr'

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return acescct_to_acescg(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return acescg_to_acescct(coords)



"""Test LCh library."""
import unittest
from . import util
from coloraide import Color, NaN
import pytest


class TestLCh(util.ColorAssertsPyTest):
    """Test LCh."""

    COLORS = [
        ('red', 'color(--lch 54.291 106.84 40.858)'),
        ('orange', 'color(--lch 75.59 83.769 70.824)'),
        ('yellow', 'color(--lch 97.607 94.712 99.572)'),
        ('green', 'color(--lch 46.278 67.984 134.38)'),
        ('blue', 'color(--lch 29.568 131.2 301.36)'),
        ('indigo', 'color(--lch 19.715 71.818 310.91)'),
        ('violet', 'color(--lch 69.618 64.617 325.57)'),
        ('white', 'color(--lch 100 0 none)'),
        ('gray', 'color(--lch 53.585 0 none)'),
        ('black', 'color(--lch 0 0 none)'),
        # Test CSS
        ('lch(100 30 270)', 'color(--lch 100 30 270 / 1)'),
        ('lch(100 30 270 / 0.5)', 'color(--lch 100 30 270 / 0.5)'),
        ('lch(50% 30 270)', 'color(--lch 50 30 270 / 1)'),
        ('lch(50% 50% 270 / 50%)', 'color(--lch 50 75 270 / 0.5)'),
        ('lch(none none none / none)', 'color(--lch none none none / none)'),
        ('lch(1 30 50%)', None),
        ('lch(1, 30, 50)', None),
        ('lch(1 30)', None),
        ('lch(1deg 30 270)', None),
        ('lch(1 30 270 50%)', None),
        # Test degrees
        ('lch(75 20 180deg)', 'color(--lch 75 20 180 / 1)'),
        ('lch(75 20 0.5turn)', 'color(--lch 75 20 180 / 1)'),
        ('lch(75 20 3.14159rad)', 'color(--lch 75 20 180 / 1)'),
        ('lch(75 20 200grad)', 'color(--lch 75 20 180 / 1)'),
        # Test color
        ('color(--lch 100 30 270)', 'color(--lch 100 30 270)'),
        ('color(--lch 100 30 270 / 0.5)', 'color(--lch 100 30 270 / 0.5)'),
        ('color(--lch 50% 50% 50% / 50%)', 'color(--lch 50 75 180 / 0.5)'),
        ('color(--lch none none none / none)', 'color(--lch none none none / none)'),
        # Test percent ranges
        ('color(--lch 0% 0% 0%)', 'color(--lch 0 0 none)'),
        ('color(--lch 100% 100% 100%)', 'color(--lch 100 150 360 / 1)'),
        ('color(--lch -100% -100% -100%)', 'color(--lch -100 0 -360 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        if color2 is None:
            with pytest.raises(ValueError):
                Color(color1)
        else:
            self.assertColorEqual(Color(color1).convert('lch'), Color(color2), color=True)


class TestLChSerialize(util.ColorAssertsPyTest):
    """Test LCh serialization."""

    COLORS = [
        # Test hex no options
        ('lch(75 30 50)', {}, 'lch(75 30 50)'),
        # Test alpha
        ('lch(75 30 50 / 0.5)', {}, 'lch(75 30 50 / 0.5)'),
        ('lch(75 30 50)', {'alpha': True}, 'lch(75 30 50 / 1)'),
        ('lch(75 30 50 / 0.5)', {'alpha': False}, 'lch(75 30 50)'),
        # Test percent
        ('lch(50 75 180)', {'percent': True}, 'lch(50% 50% 180)'),
        ('lch(50 75 180 / 0.5)', {'percent': True, 'alpha': True}, 'lch(50% 50% 180 / 0.5)'),
        # Test None
        ('lch(none 30 50)', {}, 'lch(0 30 50)'),
        ('lch(none 30 50)', {'none': True}, 'lch(none 30 50)'),
        # Test fit (not bound)
        ('lch(20 160 0)', {}, 'lch(20 160 0)'),
        ('lch(20 160 0)', {'fit': False}, 'lch(20 160 0)'),
        # Test color
        ('lch(none 30 50 / 0.5)', {'color': True}, 'color(--lch 0 30 50 / 0.5)'),
        ('lch(none 30 50)', {'color': True, 'none': True}, 'color(--lch none 30 50)'),
        ('lch(0 30 50)', {'color': True, 'alpha': True}, 'color(--lch 0 30 50 / 1)'),
        ('lch(0 30 50 / 0.5)', {'color': True, 'alpha': False}, 'color(--lch 0 30 50)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestLChProperties(util.ColorAsserts, unittest.TestCase):
    """Test LCh."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c['lightness'], 90)
        c['lightness'] = 80
        self.assertEqual(c['lightness'], 80)

    def test_chroma(self):
        """Test `chroma`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c['chroma'], 50)
        c['chroma'] = 40
        self.assertEqual(c['chroma'], 40)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c['hue'], 120)
        c['hue'] = 110
        self.assertEqual(c['hue'], 110)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('lch', [90, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('lch(90% 0 none / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_near_zero_null(self):
        """
        Test very near zero null.

        This is a fix up to help give more sane hues
        when chroma is very close to zero.
        """

        c = Color('lch(90% 0.000000000009 120 / 1)').convert('lab').convert('lch')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_chroma(self):
        """Test minimum saturation."""

        c = Color('lch(90% 0 120 / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_from_lab(self):
        """Test null from Lab conversion."""

        c1 = Color('lab(90% 0 0)')
        c2 = c1.convert('lch')
        self.assertColorEqual(c2, Color('lch(90% 0 0)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to LCh with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('lch')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to lab."""

        self.assertColorEqual(Color('lch(90% -10 120 / 1)').convert('lab'), Color('lab(90% 0 0)'))


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('lch', [30, 0, 270]).is_achromatic(), True)
        self.assertEqual(Color('lch', [30, 0.000001, 270]).is_achromatic(), True)
        self.assertEqual(Color('lch', [NaN, 0.00001, 270]).is_achromatic(), True)
        self.assertEqual(Color('lch', [0, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('lch', [0, 100, 270]).is_achromatic(), False)
        self.assertEqual(Color('lch', [NaN, 20, 270]).is_achromatic(), False)
        self.assertEqual(Color('lch', [30, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('lch', [NaN, NaN, 270]).is_achromatic(), True)



"""
Prismatic color space.

Creates a Maxwell color triangle with a lightness component.

http://psgraphics.blogspot.com/2015/10/prismatic-color-model.html
https://studylib.net/doc/14656976/the-prismatic-color-space-for-rgb-computations
"""
from __future__ import annotations
from ..spaces import Space
from ..channels import Channel
from ..cat import WHITES
from ..types import Vector
from .. import algebra as alg
import math


def srgb_to_lrgb(rgb: Vector) -> Vector:
    """Convert sRGB to Prismatic."""

    l = max(rgb)
    s = sum(rgb)
    return [l] + ([(c / s) for c in rgb] if s != 0 else [0, 0, 0])


def lrgb_to_srgb(lrgb: Vector) -> Vector:
    """Convert Prismatic to sRGB."""

    rgb = lrgb[1:]
    l = lrgb[0]
    mx = max(rgb)
    return [(l * c) / mx for c in rgb] if mx != 0 else [0, 0, 0]


class Prismatic(Space):
    """The Prismatic color class."""

    BASE = "srgb"
    NAME = "prismatic"
    SERIALIZE = ("--prismatic",)  # type: tuple[str, ...]
    EXTENDED_RANGE = False
    CHANNELS = (
        Channel("l", 0.0, 1.0, bound=True),
        Channel("r", 0.0, 1.0, bound=True),
        Channel("g", 0.0, 1.0, bound=True),
        Channel("b", 0.0, 1.0, bound=True)
    )
    CHANNEL_ALIASES = {
        "lightness": 'l',
        "red": 'r',
        "green": 'g',
        "blue": 'b'
    }
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Test if color is achromatic."""

        if math.isclose(0.0, coords[0], abs_tol=1e-4):
            return True

        white = [1, 1, 1]
        for x in alg.vcross(coords[:-1], white):
            if not math.isclose(0.0, x, abs_tol=1e-5):
                return False
        return True

    def to_base(self, coords: Vector) -> Vector:
        """To sRGB."""

        return lrgb_to_srgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From sRGB."""

        return srgb_to_lrgb(coords)



"""Test contrast."""
import unittest
from coloraide.everything import ColorAll as Color
from . import util


class TestLuminance(util.ColorAsserts, unittest.TestCase):
    """Test luminance cases."""

    def test_luminance(self):
        """Test luminance."""

        c1 = Color('orange')
        c2 = c1.convert('xyz-d65')
        self.assertEqual(c1.luminance(), c2.get('y'))

    def test_luminance_auto(self):
        """Test luminance auto white point."""

        c1 = Color('orange').convert('prophoto-rgb')
        self.assertCompare(c1.luminance(), c1.convert('xyz-d65').get('y'))
        self.assertCompare(c1.luminance(white=None), c1.convert('xyz-d50').get('y'))


class TestContrast(util.ColorAsserts, unittest.TestCase):
    """Test miscellaneous API features."""

    def test_bad_indirect_input(self):
        """Test bad input when it is done indirectly through a method."""

        with self.assertRaises(TypeError):
            Color("red").contrast(3)

    def test_contrast_dict(self):
        """Test contrast with a dictionary mapping."""

        self.assertEqual(
            Color('white').contrast('blue'),
            Color('white').contrast({"space": "srgb", "coords": [0, 0, 1]})
        )

    def test_bad_method(self):
        """Test bad contrast method."""

        with self.assertRaises(ValueError):
            Color('white').contrast('blue', method='bad')


class TestContrastWCAG21(util.ColorAsserts, unittest.TestCase):
    """Test WCAG 2.1 contrast ration specifics."""

    def test_contrast_one_wcag21(self):
        """Test contrast of one."""

        self.assertEqual(Color('blue').contrast('blue', method='wcag21'), 1)

    def test_contrast_bigger_wcag21(self):
        """Test greater contrast."""

        self.assertCompare(Color('white').contrast('blue', method='wcag21'), 8.59301)

    def test_symmetry(self):
        """Test symmetry."""

        self.assertEqual(
            Color('white').contrast('blue', method='wcag21'),
            Color('blue').contrast('white', method='wcag21'),
        )


class TestContrastLstar(util.ColorAsserts, unittest.TestCase):
    """Test L* contrast difference."""

    def test_contrast_same(self):
        """Test contrast of to same colors."""

        self.assertEqual(Color('blue').contrast('blue', method='lstar'), 0)

    def test_contrast_bigger(self):
        """Test greater contrast."""

        self.assertCompare(Color('orange').contrast('blue', method='lstar'), 42.63303)

    def test_symmetry(self):
        """Test symmetry."""

        self.assertEqual(
            Color('orange').contrast('blue', method='lstar'),
            Color('blue').contrast('orange', method='lstar')
        )



"""Test Rec.2020 library."""
import unittest
from . import util
from coloraide import Color
import pytest


class TestRec2020(util.ColorAssertsPyTest):
    """Test Rec. 2020."""

    COLORS = [
        ('red', 'color(rec2020 0.79198 0.23098 0.07376)'),
        ('orange', 'color(rec2020 0.86727 0.64078 0.18496)'),
        ('yellow', 'color(rec2020 0.97831 0.99436 0.29839)'),
        ('green', 'color(rec2020 0.23521 0.43171 0.08543)'),
        ('blue', 'color(rec2020 0.16837 0.05113 0.94678)'),
        ('indigo', 'color(rec2020 0.19583 0.03329 0.43481)'),
        ('violet', 'color(rec2020 0.80438 0.51467 0.89463)'),
        ('white', 'color(rec2020 1 1 1)'),
        ('gray', 'color(rec2020 0.45214 0.45214 0.45214)'),
        ('black', 'color(rec2020 0 0 0)'),
        # Test CSS color
        ('color(rec2020 0 0.50196 0)', 'color(rec2020 0 0.50196 0)'),
        ('color(rec2020 0 0.50196 0 / 0.5)', 'color(rec2020 0 0.50196 0 / 0.5)'),
        ('color(rec2020 50% 50% 50% / 50%)', 'color(rec2020 0.5 0.5 0.5 / 0.5)'),
        ('color(rec2020 none none none / none)', 'color(rec2020 none none none / none)'),
        # Test range
        ('color(rec2020 0% 0% 0%)', 'color(rec2020 0 0 0)'),
        ('color(rec2020 100% 100% 100%)', 'color(rec2020 1 1 1)'),
        ('color(rec2020 -100% -100% -100%)', 'color(rec2020 -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('rec2020'), Color(color2))


class TestRec2020Serialize(util.ColorAssertsPyTest):
    """Test Rec. 2020 serialization."""

    COLORS = [
        # Test color
        ('color(rec2020 0 0.3 0.75 / 0.5)', {}, 'color(rec2020 0 0.3 0.75 / 0.5)'),
        # Test alpha
        ('color(rec2020 0 0.3 0.75)', {'alpha': True}, 'color(rec2020 0 0.3 0.75 / 1)'),
        ('color(rec2020 0 0.3 0.75 / 0.5)', {'alpha': False}, 'color(rec2020 0 0.3 0.75)'),
        # Test None
        ('color(rec2020 none 0.3 0.75)', {}, 'color(rec2020 0 0.3 0.75)'),
        ('color(rec2020 none 0.3 0.75)', {'none': True}, 'color(rec2020 none 0.3 0.75)'),
        # Test Fit
        ('color(rec2020 1.2 0.2 0)', {}, 'color(rec2020 1 0.41836 0.13834)'),
        ('color(rec2020 1.2 0.2 0)', {'fit': False}, 'color(rec2020 1.2 0.2 0)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestRec2020Properties(util.ColorAsserts, unittest.TestCase):
    """Test Rec2020."""

    def test_red(self):
        """Test `red`."""

        c = Color('color(rec2020 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['red'], 0.1)
        c['red'] = 0.2
        self.assertEqual(c['red'], 0.2)

    def test_green(self):
        """Test `green`."""

        c = Color('color(rec2020 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['green'], 0.2)
        c['green'] = 0.1
        self.assertEqual(c['green'], 0.1)

    def test_blue(self):
        """Test `blue`."""

        c = Color('color(rec2020 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['blue'], 0.3)
        c['blue'] = 0.1
        self.assertEqual(c['blue'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(rec2020 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Test Averaging."""
import unittest
from coloraide.everything import ColorAll as Color
from . import util
import math


class TestAverage(util.ColorAsserts, unittest.TestCase):
    """Test averaging."""

    def test_no_colors(self):
        """Test averaging no colors."""

        with self.assertRaises(ValueError):
            Color.average([])

    def test_one_colors(self):
        """Test averaging one color."""

        self.assertColorEqual(Color.average(['red']), Color('color(srgb-linear 1 0 0)'))

    def test_average_premultiplied(self):
        """Test averaging with premultiplication."""

        colors = [Color('darkgreen'), Color('color(srgb 0 0.50196 0 / 1)'), Color('color(srgb 0 0 1)')]
        results = []
        for i in range(5):
            colors[1].set('alpha', i / 4)
            results.append(Color.average(colors, space='srgb').to_string(color=True))
        self.assertEqual(
            results,
            ['color(srgb 0 0.19608 0.5 / 0.66667)',
             'color(srgb 0 0.23007 0.44444 / 0.75)',
             'color(srgb 0 0.25725 0.4 / 0.83333)',
             'color(srgb 0 0.2795 0.36364 / 0.91667)',
             'color(srgb 0 0.29804 0.33333)']
        )

    def test_average_no_premultiplied(self):
        """Test averaging without premultiplication."""

        colors = [Color('darkgreen'), Color('color(srgb 0 0.50196 0 / 1)'), Color('color(srgb 0 0 1)')]
        results = []
        for i in range(5):
            colors[1].set('alpha', i / 4)
            results.append(Color.average(colors, space='srgb', premultiplied=False).to_string(color=True))
        self.assertEqual(
            results,
            ['color(srgb 0 0.29804 0.33333 / 0.66667)',
             'color(srgb 0 0.29804 0.33333 / 0.75)',
             'color(srgb 0 0.29804 0.33333 / 0.83333)',
             'color(srgb 0 0.29804 0.33333 / 0.91667)',
             'color(srgb 0 0.29804 0.33333)']
        )

    def test_average_cylindrical_premultiplied(self):
        """Test cylindrical averaging without premultiplication."""

        colors = [Color('darkgreen'), Color('color(srgb 0 0.50196 0 / 1)'), Color('color(srgb 0 0 1)')]
        results = []
        for i in range(5):
            colors[1].set('alpha', i / 4)
            results.append(Color.average(colors, space='hsl').to_string(color=True))
        self.assertEqual(
            results,
            ['color(--hsl 150 1 0.34804 / 0.66667)',
             'color(--hsl 150 1 0.33725 / 0.75)',
             'color(--hsl 150 1 0.32863 / 0.83333)',
             'color(--hsl 150 1 0.32157 / 0.91667)',
             'color(--hsl 150 1 0.31569)']
        )

    def test_average_cylindrical_no_premultiplied(self):
        """Test cylindrical averaging without premultiplication."""

        colors = [Color('darkgreen'), Color('color(srgb 0 0.50196 0 / 1)'), Color('color(srgb 0 0 1)')]
        results = []
        for i in range(5):
            colors[1].set('alpha', i / 4)
            results.append(Color.average(colors, space='hsl', premultiplied=False).to_string(color=True))
        self.assertEqual(
            results,
            ['color(--hsl 150 1 0.31569 / 0.66667)',
             'color(--hsl 150 1 0.31569 / 0.75)',
             'color(--hsl 150 1 0.31569 / 0.83333)',
             'color(--hsl 150 1 0.31569 / 0.91667)',
             'color(--hsl 150 1 0.31569)']
        )

    def test_average_ignore_undefined(self):
        """Test averaging ignores undefined."""

        colors = [Color('darkgreen'), Color('color(srgb 0 none 0 / 1)'), Color('color(srgb 0 0 1)')]
        results = []
        for i in range(5):
            colors[1].set('alpha', i / 4)
            results.append(Color.average(colors, space='srgb').to_string(color=True))
        self.assertEqual(
            results,
            ['color(srgb 0 0.29412 0.5 / 0.66667)',
             'color(srgb 0 0.26144 0.44444 / 0.75)',
             'color(srgb 0 0.23529 0.4 / 0.83333)',
             'color(srgb 0 0.2139 0.36364 / 0.91667)',
             'color(srgb 0 0.19608 0.33333)']
        )

    def test_average_ignore_undefined_alpha_premultiplied(self):
        """Test averaging ignores undefined."""

        colors = [Color('darkgreen'), Color('color(srgb 0 0.50196 0 / none)'), Color('color(srgb 0 0 1)')]
        self.assertEqual(Color.average(colors, space='srgb').to_string(color=True), 'color(srgb 0 0.29804 0.33333)')

    def test_all_undefined_hue(self):
        """Test all hues undefined."""

        a = Color.average(['white', 'gray', 'black'], space='hsl')
        self.assertColorEqual(a, Color('hsl(none 0% 50.065%)'))
        self.assertTrue(math.isnan(a[0]))

    def test_all_undefined_non_hue(self):
        """Test all non hue undefined."""

        a = Color.average(['color(srgb 1 none 0)', 'color(srgb 0.1 none 0.3)', 'color(srgb 0.7 none 0)'], space='srgb')
        self.assertColorEqual(a, Color('rgb(153 none 25.5)'))
        self.assertTrue(math.isnan(a[1]))



"""
The IPT color space.

https://www.researchgate.net/publication/\
221677980_Development_and_Testing_of_a_Color_Space_IPT_with_Improved_Hue_Uniformity.
"""
from __future__ import annotations
from ..spaces import Space, Labish
from ..channels import Channel, FLG_MIRROR_PERCENT
from .. import algebra as alg
from .achromatic import Achromatic as _Achromatic
from .srgb_linear import lin_srgb_to_xyz
from .srgb import lin_srgb
from ..types import Vector
from typing import Any
import math
from .. import util

XYZ_TO_LMS = [
    [0.4002, 0.7075, -0.0807],
    [-0.2280, 1.1500, 0.0612],
    [0.0, 0.0, 0.9184]
]

LMS_TO_XYZ = [
    [1.8502429449432056, -1.1383016378672328, 0.23843495850870136],
    [0.36683077517134854, 0.6438845448402355, -0.010673443584379992],
    [0.0, 0.0, 1.088850174216028]
]

LMS_P_TO_IPT = [
    [0.4, 0.4, 0.2],
    [4.455, -4.851, 0.396],
    [0.8056, 0.3572, -1.1628]
]

IPT_TO_LMS_P = [
    [1.0000000000000004, 0.0975689305146139, 0.2052264331645916],
    [0.9999999999999997, -0.1138764854731471, 0.13321715836999803],
    [1.0, 0.0326151099170664, -0.6768871830691793]
]

ACHROMATIC_RESPONSE = [
    [0.017066845239980113, 1.321844777655898e-06, 329.76026731764046],
    [0.022993026958471587, 1.7808336678784566e-06, 329.76026731797435],
    [0.02737255832988907, 2.1200328924793134e-06, 329.7602673179663],
    [0.03097697795223092, 2.3991989122003048e-06, 329.7602673180789],
    [0.9999910919149724, 7.745034210925492e-05, 329.7602673179579],
    [5.243613106559707, 0.0004061232467725615, 329.7602673179984]]  # type: list[Vector]


def xyz_to_ipt(xyz: Vector) -> Vector:
    """XYZ to IPT."""

    lms_p = [alg.npow(c, 0.43) for c in alg.dot(XYZ_TO_LMS, xyz, dims=alg.D2_D1)]
    return alg.dot(LMS_P_TO_IPT, lms_p, dims=alg.D2_D1)


def ipt_to_xyz(ipt: Vector) -> Vector:
    """IPT to XYZ."""

    lms = [alg.nth_root(c, 0.43) for c in alg.dot(IPT_TO_LMS_P, ipt, dims=alg.D2_D1)]
    return alg.dot(LMS_TO_XYZ, lms, dims=alg.D2_D1)


class Achromatic(_Achromatic):
    """Test achromatic response."""

    def convert(self, coords: Vector, **kwargs: Any) -> Vector:
        """Convert to the target color space."""

        lab = xyz_to_ipt(lin_srgb_to_xyz(lin_srgb(coords)))
        l = lab[0]
        c, h = alg.rect_to_polar(*lab[1:])
        return [l, c, h]


class IPT(Labish, Space):
    """The IPT class."""

    BASE = "xyz-d65"
    NAME = "ipt"
    SERIALIZE = ("--ipt",)  # type: tuple[str, ...]
    CHANNELS = (
        Channel("i", 0.0, 1.0),
        Channel("p", -1.0, 1.0, flags=FLG_MIRROR_PERCENT),
        Channel("t", -1.0, 1.0, flags=FLG_MIRROR_PERCENT)
    )
    CHANNEL_ALIASES = {
        "intensity": "i",
        "protan": "p",
        "tritan": "t"
    }

    # The D65 white point used in the paper was different than what we use.
    # We use chromaticity points (0.31270, 0.3290) which gives us an XYZ of ~[0.9505, 1.0000, 1.0890]
    # IPT uses XYZ of [0.9504, 1.0, 1.0889] which yields chromaticity points ~(0.3127035830618893, 0.32902313032606195)
    WHITE = tuple(util.xyz_to_xyY([0.9504, 1.0, 1.0889])[:-1])  # type: ignore[assignment]
    # Precalculated from:
    # [
    #     (1, 5, 1, 1000.0),
    #     (100, 101, 1, 100),
    #     (520, 521, 1, 100)
    # ]
    ACHROMATIC = Achromatic(
        ACHROMATIC_RESPONSE,
        1e-5,
        1e-5,
        0.00049,
        'linear',
        mirror=True
    )  # type: _Achromatic

    def resolve_channel(self, index: int, coords: Vector) -> float:
        """Resolve channels."""

        if index in (1, 2):
            if not math.isnan(coords[index]):
                return coords[index]

            return self.ACHROMATIC.get_ideal_ab(coords[0])[index - 1]

        value = coords[index]
        return self.channels[index].nans if math.isnan(value) else value

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        m, h = alg.rect_to_polar(coords[1], coords[2])
        return self.ACHROMATIC.test(coords[0], m, h)

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return ipt_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return xyz_to_ipt(coords)



"""
Delta E 99o.

https://de.wikipedia.org/wiki/DIN99-Farbraum
"""
from __future__ import annotations
from .delta_e_76 import DE76


class DE99o(DE76):
    """Delta E 99o class."""

    NAME = "99o"
    SPACE = "din99o"



"""Get the average color of an image."""
from functools import lru_cache
from PIL import Image
import time
import argparse
import sys
import os

# We want to load ColorAide from the working directory to pick up the version under development
sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color


def printt(t):
    """Print time."""

    print('Completed in: ', end='')
    s = t // 1e+9
    m = t // 1e+6
    u = t // 1000
    if s:
        s = t / 1e+9
        h = m = 0
        m = s // 60
        if m:
            s -= m * 60
            h = m // 60
            if h:
                m -= h * 60
        if h:
            print('{} hours '.format(h), end='')
        if m:
            print('{} minutes '.format(m), end='')
        print('{} sec'.format(s))
    elif m:
        print('{} msec'.format(t / 1e+6))
    elif u:
        print('{} usec'.format(t / 1000))
    else:
        print('{} nsec'.format(t))


@lru_cache(maxsize=1024 * 1024)
def get_color(space, p):
    """Translate the pixel to a color."""

    return Color('srgb', [x / 255 for x in p[:3]], p[3] / 255 if len(p) > 3 else 1).convert(space, in_place=True)


def iter_image(img, space):
    """Process the image applying the requested filter."""

    with Image.open(img) as im:
        pixels = im.load()
        start = time.perf_counter_ns()
        total = im.size[0] * im.size[1]
        factor = 100 / total
        i = j = 0
        print('Pixels: {}'.format(total))
        print('> 0%', end='\r')
        for i in range(im.size[0]):
            for j in range(im.size[1]):
                yield get_color(space, pixels[i, j])
            print('> {}%'.format(int((i * j) * factor)), end="\r")
        print('> 100%')
        t = time.perf_counter_ns() - start
        printt(t)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='average_img', description='Get the average color of an image.')
    parser.add_argument('--input', '-i', help='Input image.')
    parser.add_argument('--space', '-s', default='srgb', help='Color space to average in.')
    parser.add_argument('--out-space', '-o', default='srgb', help='Color space to average in.')
    parser.add_argument('--premultiplied', '-p', action='store_true', help="Premultiply values.")
    args = parser.parse_args()

    print(
        Color.average(
            iter_image(args.input, args.space),
            space=args.space,
            premultiplied=args.premultiplied
        ).convert(args.out_space).to_string()
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
The xyY color space.

https://en.wikipedia.org/wiki/CIE_1931_color_space#CIE_xy_chromaticity_diagram_and_the_CIE_xyY_color_space
"""
from __future__ import annotations
from ..spaces import Space
from ..channels import Channel
from ..cat import WHITES
from .. import util
from ..types import Vector
from .. import algebra as alg
import math


class xyY(Space):
    """The xyY class."""

    BASE = "xyz-d65"
    NAME = "xyy"
    SERIALIZE = ("--xyy",)
    CHANNELS = (
        Channel("x", 0.0, 1.0),
        Channel("y", 0.0, 1.0),
        Channel("Y", 0.0, 1.0)
    )
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Test if color is achromatic."""

        if math.isclose(0.0, coords[-1], abs_tol=1e-4):
            return True

        for x in alg.vcross(coords[:-1], self.WHITE):
            if not math.isclose(0.0, x, abs_tol=1e-6):
                return False
        return True

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return util.xy_to_xyz(coords[0:2], coords[2])

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return util.xyz_to_xyY(coords, self.white())



"""sRGB Linear color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector


RGB_TO_XYZ = [
    [0.41239079926595923, 0.35758433938387807, 0.1804807884018343],
    [0.21263900587151022, 0.7151686787677561, 0.07219231536073371],
    [0.019330818715591818, 0.11919477979462599, 0.9505321522496607]
]

XYZ_TO_RGB = [
    [3.240969941904524, -1.5373831775700946, -0.4986107602930036],
    [-0.9692436362808795, 1.8759675015077202, 0.04155505740717561],
    [0.05563007969699365, -0.20397695888897652, 1.0569715142428784]
]


def lin_srgb_to_xyz(rgb: Vector) -> Vector:
    """
    Convert an array of linear-light sRGB values to CIE XYZ using sRGB's own white.

    D65 (no chromatic adaptation)
    """

    return alg.dot(RGB_TO_XYZ, rgb, dims=alg.D2_D1)


def xyz_to_lin_srgb(xyz: Vector) -> Vector:
    """Convert XYZ to linear-light sRGB."""

    return alg.dot(XYZ_TO_RGB, xyz, dims=alg.D2_D1)


class sRGBLinear(sRGB):
    """sRGB linear."""

    BASE = 'xyz-d65'
    NAME = "srgb-linear"
    SERIALIZE = ("srgb-linear",)
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from sRGB Linear."""

        return lin_srgb_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to sRGB Linear."""

        return xyz_to_lin_srgb(coords)



"""
Generate random points in and around the Pointer gamut indicate whether they are inside or outside the gamut.

If a color is given, plot it against the gamut instead of random points.
"""
import sys
import os
import math
import random
import argparse
import matplotlib.pyplot as plt

sys.path.insert(0, os.getcwd())

from coloraide import gamut  # noqa: E402
from coloraide import algebra as alg  # noqa: E402
from coloraide.everything import ColorAll  # noqa: E402
from coloraide.spaces.lch import LCh  # noqa: E402
from coloraide.spaces.lab import Lab  # noqa: E402


class LabPointer(Lab):
    """Lab Pointer's Gamut."""

    BASE = 'xyz-d65'
    NAME = 'lab-pointer'
    SERIALIZE = ('--lab-pointer',)
    WHITE = gamut.pointer.WHITE_POINT_SC


class LChPointer(LCh):
    """LCh Pointer's Gamut."""

    BASE = 'lab-pointer'
    NAME = 'lch-pointer'
    SERIALIZE = ('--lch-pointer',)
    WHITE = gamut.pointer.WHITE_POINT_SC


class Color(ColorAll):
    """Custom class for Pointer conversion."""


Color.register([LabPointer(), LChPointer()])


def plot_pointer_gamut(target, space_gamut, fit, title, dark=False):
    """
    Plot the Pointer gamut and random points inside and outside the gamut.

    If a color is given, plot it against the gamut instead of random points.
    """

    if not dark:
        plt.style.use('seaborn-v0_8-darkgrid')
        default_color = 'black'
    else:
        plt.style.use('dark_background')
        default_color = 'white'

    lightness = random.uniform(15, 90)
    colors = []
    if not target:
        for _ in range(20):
            while True:
                color = Color.random('lch-pointer', limits=[(lightness, lightness), None, None])
                if color.in_gamut(space_gamut, tolerance=0):
                    colors.append(color)
                    break
    else:
        original = Color(target)
        colors.append(original.convert('lch-pointer', norm=False))
        lightness = colors[0]['l']
        if fit and not colors[0].in_pointer_gamut():
            colors.append(colors[0].clone().fit_pointer_gamut())
    color = colors[0]
    l, c, h = color[:-1]
    li, lf = gamut.pointer.closest_lightness(l)
    chroma = [alg.lerp(row[li], row[li + 1], lf) for row in gamut.pointer.LCH_POINTER]
    chroma.append(chroma[0])

    figure = plt.figure()

    ax = plt.axes(
        xlabel='hue',
        ylabel='chroma',
        projection='polar'
    )

    ax.set_aspect('equal')
    figure.add_axes(ax)

    if not title:
        plt.suptitle('Pointer Gamut: LCh SC at Lightness {}'.format(round(lightness, 2)))
    if target:
        if len(colors) > 1:
            space = original.space()
            ax.set_title(
                'Color: {} -> {}'.format(colors[0].convert(space).to_string(), colors[1].convert(space).to_string()),
                fontdict={'fontsize': 8},
                pad=2
            )
        else:
            ax.set_title('Color: {}'.format(colors[0].to_string()), fontdict={'fontsize': 8}, pad=2)

    plt.plot(
        [math.radians(hue) for hue in gamut.pointer.LCH_H] + [0.0],
        chroma,
        color=default_color,
        marker="",
        linewidth=1.5,
        markersize=2,
        antialiased=True
    )

    inside = []
    outside = []

    for color in colors:
        c, h = color[1:-1]
        if color.in_pointer_gamut():
            inside.append([c, math.radians(h)])
        else:
            outside.append([c, math.radians(h)])

    if inside:
        y, x = zip(*inside)
        plt.scatter(
            x,
            y,
            marker="o",
            color=default_color,
            edgecolor=default_color,
            s=16,
            zorder=100
        )

    if outside:
        y, x = zip(*outside)
        plt.scatter(
            x,
            y,
            marker="x",
            color=default_color,
            s=16,
            zorder=100
        )


def main():
    """Main."""

    parser = argparse.ArgumentParser(
        prog='pointer_gamut_diagram', description='Plot random points in relation to Pointer gamut.'
    )
    parser.add_argument(
        '--gamut', '-g', default="display-p3", help='Gamut to evaluate the color in (default is display-p3).'
    )
    parser.add_argument('--fit', '-f', action="store_true", help="Fit color to pointer gamut.")
    parser.add_argument(
        '--color', '-c',
        help="Color to test against the Pointer gamut, single colors will show the original and fitted color."
    )
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()

    plot_pointer_gamut(
        args.color,
        args.gamut,
        args.fit,
        args.title,
        dark=args.dark
    )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())
